import os,sys,glob,json,shutil,copy
from abacustest.lib_prepare.abacus import AbacusStru,ReadInput,WriteInput,WriteKpt
from abacustest.lib_collectdata.collectdata import RESULT
from abacustest.prepare import PrepareAbacus
from .model import Model
from . import comm
from abacustest.lib_prepare.comm import kspacing2kpt

import matplotlib.pyplot as plt
import numpy as np
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE

class fdstress(Model):
    '''
    Base class for all models
    
    Some specific workflow will be implemented in the subclass.
    Such as EOS, BAND, etc.
    
    The model should have the following methods:
    '''
    @staticmethod
    def add_args(parser):
        '''
        Add arguments for the model
        The arguments can not be command, model, modelcommand 
        '''
        pass
    
    def run(self,params):
        '''
        Parse the parameters and run the model
        '''
        pass
    
    @staticmethod
    def prepare_args(parser):
        '''
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand '''
        
        parser.description = "Do the finite difference of stress test."
        parser.add_argument('-d', '--step', type=float, default=0.0001,help="the step to change the cell, default 0.0001. This is a coefficient, the real step is step*cell_vector")
        parser.add_argument('-n', '--number', type=int,  default=5,help='the number of points to calculate the stress tensor, default 5')
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus inputs, default is current folder.')
        parser.add_argument('--comp',type=str,default=["11","12","13","22","23","33"], nargs="*" ,help='the stress components to test, default is 11 12 13 22 23 33.')
        return parser
    
    def run_prepare(self,params):
        '''
        Parse the parameters and run the prepare process.
        Usually, this step will generate the input files for abacustest submit.
        '''
        jobs = ["."] if len(params.job) == 1 else params.job[1:]

        subfolders = PrepareFDStress(jobs,params.step,params.number,params.comp).run()

        if subfolders:
            setting = {
                "save_path": "results",
                "bohrium_group_name": "FDStress",
                "run_dft": {
                    "example": subfolders,
                    "command": RECOMMAND_COMMAND,
                    "image": RECOMMAND_IMAGE,
                    "bohrium": {
                        "scass_type": RECOMMAND_MACHINE,
                        "job_type": "container",
                        "platform": "ali",
                    },
                },
            }
            comm.dump_setting(setting)
            comm.doc_after_prepare("FDStress", subfolders, ["setting.json"],has_prepare=False)
    
    @staticmethod
    def postprocess_args(parser):
        '''
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand'''
        parser.description = "Postprocess the finite difference of stress test. There should have several cell_* folders in the job folder."
        parser.add_argument('-j', '--job',default=["."], action="extend",nargs="*" ,help='the path of abacus results, default is current folder.')
        parser.add_argument('--read',type=int, const=1, default=0,nargs="?",help='if read the results.json in each job folder, default is 0, no read.')
        parser.add_argument('-t', '--type',default=0,type=int ,help='the job type. 0:abaus, 1:qe. Default 0')
        return parser

    def run_postprocess(self,params):
        '''
        Parse the parameters and run the postprocess process'''
        jobs = ["."] if len(params.job) == 1 else params.job[1:]
        PostProcessFDStress(jobs,params.read,params.type).run()
    
class PrepareFDStress:
    def __init__(self,jobs,step,number,comp=["11","12","13","22","23","33"]):
        self.jobs = jobs
        self.step = step
        self.number = number
        self.comp = comp
    
    def run(self):
        print("prepare inputs")
        print("step:",self.step)
        print("number:",self.number)
        subfolders = []
        for job in self.jobs:
            for ijob in glob.glob(job):
                # remove fdf folders
                for ifile in glob.glob(os.path.join(ijob,"cell_*")):
                    if os.path.isdir(ifile):
                        shutil.rmtree(ifile)
                print("prepare inputs in ",ijob)
                ifolders = self.preapre_stru(ijob)
                if ifolders:
                    subfolders.extend(ifolders)
        return subfolders
    
    def gen_kpt(self, init_path, kspacing, cell):
        kpt = kspacing2kpt(kspacing, cell)
                
    def preapre_stru(self,init_path):
        input_param = {"calculation": "scf","cal_stress":1}
        otherfiles = []
        for ifile in os.listdir(init_path):
            if ifile == "INPUT":
                input_param = PrepareAbacus.ReadInput(os.path.join(init_path,"INPUT"))
                input_param["calculation"] = "scf"
                input_param["cal_stress"] = 1
                continue
            elif ifile != "STRU" and not ifile.startswith(".") and os.path.isfile(os.path.join(init_path,ifile)):
                otherfiles.append(os.path.abspath(os.path.join(init_path,ifile)))
                
        stru = AbacusStru.ReadStru(os.path.join(init_path,"STRU"))
        if not stru:
            print(f"ERROR: read STRU failed in {init_path}")
            return None
                
        cell = stru.get_cell(bohr=True)  # need use cell as the input of AbacusStru
        #print(label,atom_num,pp,orb,mass,dpks,coord,cell)
        
        # we need to fix the kpt.
        kspacing = input_param.pop("kspacing",None)
        if kspacing is not None:
            kpt = kspacing2kpt(kspacing, cell)
            kptf = (os.path.abspath(os.path.join(init_path,"KPT")))
            WriteKpt(kpt,kptf)
            if kptf not in otherfiles:
                otherfiles.append(kptf)

        write_0 = False
        subfolders = []
        for i in range(3):
            for j in range(3):
                # i is the index of cell vector, j is the index of x/y/z component of cell vector
                for k in range(self.number*-1,self.number+1):
                    # k is the index of diff coef
                    if k == 0 and write_0:
                       continue
                    if f"{i+1}{j+1}" not in self.comp:
                        continue
                    sub_folder = os.path.join(init_path,f"cell_{i+1}_{j+1}_{k}")
                    if k == 0:
                        sub_folder = os.path.join(init_path,f"cell_0")
                    os.makedirs(sub_folder,exist_ok=True)
                    
                    if k == 0:
                        json.dump({"step":self.step,"number":self.number},open(os.path.join(sub_folder,"param.json"),"w"))
                        write_0 = True

                    new_cell = copy.deepcopy(cell)
                    for m in range(3):
                        new_cell[m][i] += k * self.step * cell[m][j]
                    new_stru = copy.deepcopy(stru)
                    new_stru.set_cell(new_cell,bohr=True,change_coord=True)
                    new_stru.write(os.path.join(sub_folder,"STRU"))
                    WriteInput(input_param,os.path.join(sub_folder,"INPUT"))
                    
                    # copy other files to sub_folder
                    for file in otherfiles:
                        if not os.path.isfile(os.path.join(sub_folder,os.path.basename(file))):
                            os.symlink(file,os.path.join(sub_folder,os.path.basename(file)))
                    subfolders.append(sub_folder)
        return subfolders
    
class PostProcessFDStress:
    def __init__(self,jobs,readr,jobtype=0):
        self.jobs = jobs
        self.readr = readr
        
        jobtype_dict = {0:"abacus",1:"qe"}
        if jobtype not in jobtype_dict:
            print(f"jobtype {jobtype} is not supported.")
            jobtype = 0
        self.jobtype = jobtype_dict[jobtype]
        
        unit_trans = {
            0: 1602.1757722389546,  # in abacus, the stress transfer to kbar is multiply a coef = ModuleBase::RYDBERG_SI / pow(ModuleBase::BOHR_RADIUS_SI, 3) * 1.0e-8;
                 # = 4.35974394e-18 / 2 / pow(0.529177e-10, 3) * 1.0e-8
                 # Ry to eV = 13.605698
                 # bohr3 to angstrom3 = pow(ModuleBase::BOHR_TO_A, 3) = 0.5291770**3
                 # so, when I get energy in eV and volume in angstrom3, I can get the stress = energy / 13.605698 / (volume/0.5291770**3)  * 4.35974394e-18 / 2 / pow(0.529177e-10, 3) * 1.0e-8
                 # = energy / volume * 1602.1757722389546
            1: 1602.1777023087203  # in QE, the energy and stress are transfered by abacustest
                                    # KBAR2HARTREEPERBOHR3 = 3.398927420868445E-6
                                    # BOHR2A = 0.52917721092, HARTREE2EV = 27.211396132
                                    # so, stress = energy / 27.211396132 / (volume / 0.52917721092**3) / 3.398927420868445E-6 = energy / volume * 1602.1777023087203
        }
        self.unitcoef = unit_trans[jobtype]
        
    def run(self):
        alljobs = []
        for job in self.jobs:
            for ijob in glob.glob(job):
                if os.path.isdir(ijob) and ijob not in alljobs:
                    alljobs.append(ijob) 
        sm = {}   
        metrics = {}       
        for ijob in alljobs:
            if os.path.isdir(ijob):
                print("postprocess in ",ijob)
                jsonf = os.path.join(ijob,"results.json")
                if self.readr and os.path.isfile(jsonf):
                    values = json.load(open(jsonf,"r"))
                else:
                    if not os.path.isfile(os.path.join(ijob,"cell_0","param.json")):
                        print(f"param.json is not exist in {ijob}/cell_0")
                        continue
                    param = json.load(open(os.path.join(ijob,"cell_0","param.json"),"r"))
                    values = {"param": param}
                    for cell in glob.glob(f"{ijob}/cell_*"):
                        values[os.path.basename(cell)] = self.get_value(cell)
                    json.dump(values,open(jsonf,"w"),indent=4)
                
                for k,v in values.items():
                    if k != "param":
                        metrics[os.path.join(ijob,k)] = v
                    
                if "step" not in values.get("param",{}) or "number" not in values.get("param",{}):
                    print(f"param.json is not correct in {ijob}. Need step and number.")
                    continue
                diff = values.get("param",{}).get("step",0.0001)
                number = values.get("param",{}).get("number",6)

                results1 = self.gen_result1(values,diff,number)
                results2 = self.gen_result2(values,diff,number)
                json.dump(results1,open(f"{ijob}/results1.json","w"),indent=4)
                json.dump(results2,open(f"{ijob}/results2.json","w"),indent=4)
                self.plot1(results1,f"{ijob}/stress_tensor1.png")
                self.plot2(results2,f"{ijob}/stress_tensor2.png")
                sm[f"{ijob}/stress_tensor1"] = {
                    "type": "image",
                    "file": f"{ijob}/stress_tensor1.png"
                }
                sm[f"{ijob}/stress_tensor2"] = {
                    "type": "image",
                    "file": f"{ijob}/stress_tensor2.png"
                }
                self.write2csv(ijob,results1,results2)

        # write the supermetrics.json
        json.dump(metrics,open("metrics.json","w"),indent=4)
        json.dump(sm,open("supermetrics.json","w"),indent=4)
        
    def get_value(self,ipath):
        if not os.path.isdir(ipath):
            return {}

        iresult = RESULT(fmt=self.jobtype,path=ipath)
        if iresult["stress"] != None:
            stress= [[iresult["stress"][0],iresult["stress"][1],iresult["stress"][2]],
                      [iresult["stress"][3],iresult["stress"][4],iresult["stress"][5]],
                      [iresult["stress"][6],iresult["stress"][7],iresult["stress"][8]]]
        else:
            stress = None
        return {"stress": stress,
                "energy": iresult["energy"],
                "version": iresult["version"],
                "volume": iresult["volume"],
                "converge": iresult["converge"],
                "drho_last": iresult["drho_last"],
                "denergy_last": iresult["denergy_last"],
                "total_time": iresult["total_time"]}

    def gen_result1(self,value_dict,diff,number):
        # calculate the stress tensor of -number+1, ..., number-1,
        # and compare the difference between the calculated stress tensor and the stress tensor from abacus.
        # will return a dict
        results = {
            "x":[i*diff for i in range(-number+1,number)],
            "11":{"analytical":[],"numerical":[]},
            "12":{"analytical":[],"numerical":[]},
            "13":{"analytical":[],"numerical":[]},
            "21":{"analytical":[],"numerical":[]},
            "22":{"analytical":[],"numerical":[]},
            "23":{"analytical":[],"numerical":[]},
            "31":{"analytical":[],"numerical":[]},
            "32":{"analytical":[],"numerical":[]},
            "33":{"analytical":[],"numerical":[]}
        }

        for i in range(3):
            for j in range(3):
                ij = f"{i+1}{j+1}"
                for k in range(number*-1 + 1,number):
                    cell_name = f"cell_{i+1}_{j+1}_{k}"
                    if k == 0:
                        cell_name = "cell_0"
                    volume = value_dict.get(cell_name,{}).get("volume",None)
                    stress_abacus = value_dict.get(cell_name,{}).get("stress",None)
                    if stress_abacus != None:
                        stress_abacus = stress_abacus[i][j]
                    results[ij]["analytical"].append(stress_abacus)

                    cell1 = f"cell_{i+1}_{j+1}_{k-1}"
                    cell2 = f"cell_{i+1}_{j+1}_{k+1}"

                    if k -1 == 0:
                        cell1 = "cell_0"
                    if k +1 == 0:
                        cell2 = "cell_0"
                    v1 = value_dict.get(cell1,{}).get("volume",None)
                    v2 = value_dict.get(cell2,{}).get("volume",None)
                    #print(i,j,v1,v2)
                    energy_k1 = value_dict.get(cell1,{}).get("energy",None)
                    energy_k2 = value_dict.get(cell2,{}).get("energy",None)
                    
                    stress_finite_diff = None
                    if energy_k1 == None:
                        print(f"Get energy from cell_{i+1}_{j+1}_{k-1} failed.")
                    if energy_k2 == None:
                        print(f"Get energy from cell_{i+1}_{j+1}_{k+1} failed.")
                    if volume == None:
                        print(f"Get volume from cell_{i+1}_{j+1}_{k} failed.")
                        
                    if not (energy_k1 == None or energy_k2 == None or volume == None):
                        if i == j:
                            rdiff = diff / (1 + k*diff)
                        else:
                            rdiff = diff
                        #stress_finite_diff = (energy_k1 - energy_k2) / (2*diff) / volume * 1602.17662
                        stress_finite_diff = (energy_k1 - energy_k2) / (2*rdiff) / volume * self.unitcoef

                    results[ij]["numerical"].append(stress_finite_diff)
        return results              

    def gen_result2(self,value_dict,diff,number):
        # calculate the stress tensor with diff coef 0, and based on different diff coef k and -k.
        # will return the deviation of stress tensor (numerical - analytical).
        results = {
            "x": [i*diff*2 for i in range(1,number+1)],
            "11":[],"12":[],"13":[],
            "21":[],"22":[],"23":[],
            "31":[],"32":[],"33":[],
        }
        volume = value_dict.get("cell_0",{}).get("volume",None)
        stress_analytical = value_dict.get("cell_0",{}).get("stress",None)

        for i in range(3):
            for j in range(3):
                ij = f"{i+1}{j+1}"
                for k in range(1,number+1):

                    e1 = value_dict.get(f"cell_{i+1}_{j+1}_{-k}",{}).get("energy",None)
                    e2 = value_dict.get(f"cell_{i+1}_{j+1}_{k}",{}).get("energy",None)
                    v1 = value_dict.get(f"cell_{i+1}_{j+1}_{-k}",{}).get("volume",None)
                    v2 = value_dict.get(f"cell_{i+1}_{j+1}_{k}",{}).get("volume",None)

                    if e1 == None or e2 == None or volume == None or stress_analytical == None:
                        deviation = None
                    else:
                        stress_finite_diff = (e1 - e2) / (2*k*diff) / volume * self.unitcoef
                        #stress_finite_diff = (e1 - e2) /(v2 - v1) * 1602.17662
                        deviation = stress_finite_diff - stress_analytical[i][j]

                    results[ij].append(deviation)
        return results

    def plot1(self,results,fname):
        # plot two figures, one is the both analytical and numerical stress tensor, the other is the deviation.
        # x axis is the diff coef, y axis is the stress tensor, each component of stress tensor is one legend.
        # Plot two figures in one figure, and the first figure has 9*2 legends, the second figure has 9 legends.
        fig, axs = plt.subplots(2,1,figsize=(12,10),sharex=True)
        axs[0].set_title("stress tensor",fontsize=14)
        axs[1].set_title("deviation (F.D. - Ana.))",fontsize=14)
        #axs[0].set_xlabel("diff coef")
        axs[1].set_xlabel("configurations (cell deformation degree)(F.D.=Finite Difference, Ana.=Analytical)",fontsize=14)
        axs[0].set_ylabel("stress tensor (kbar)",fontsize=14)
        axs[1].set_ylabel("deviation (kbar)",fontsize=14)
        colors = ["red","orange","purple","brown","green","pink","gray","olive","blue"]
        for i in range(3):
            for j in range(3):
                ij = f"{i+1}{j+1}"
                x,y1,y2 = comm.clean_none_list(results["x"],results[ij]["analytical"],results[ij]["numerical"])
                if len(x) > 0:
                    axs[0].plot(x,y1,label=f"Ana. {ij}",linestyle="--",color=colors[i*3+j],marker="+")
                    axs[0].plot(x,y2,label=f"F.D. {ij}",linestyle="-",color=colors[i*3+j],marker="o")
                    deviation = np.array(y1) - np.array(y2)
                    axs[1].plot(x,deviation,label=f"Dev. {ij}",linestyle="-",color=colors[i*3+j],marker="o")
        axs[0].legend(bbox_to_anchor=(1, 1), loc='upper left', borderaxespad=0.)
        axs[1].legend(bbox_to_anchor=(1, 1), loc='upper left', borderaxespad=0.)
        plt.savefig(fname)
        plt.close()
        return


    def plot2(self,results,fname):
        # plot one figure, x axis is the diff coef, y axis is the deviation of stress tensor, each component of stress tensor is one legend.
        # Plot one figure, and the figure has 9 legends.

        fig, axs = plt.subplots(1,1,figsize=(12,5))
        axs.set_title("deviation (F.D. - Ana.))",fontsize=14)
        axs.set_xlabel("finite difference step (cell deformation degree) (F.D.=Finite Difference, Ana.=Analytical)",fontsize=14)
        axs.set_ylabel("deviation (kbar)",fontsize=14)
        colors = ["red","orange","purple","brown","green","pink","gray","olive","blue"]
        for i in range(3):
            for j in range(3):
                ij = f"{i+1}{j+1}"
                x,y = comm.clean_none_list(results["x"],results[ij])
                if len(x) > 0:
                    axs.plot(x,y,label=f"Dev. {ij}",linestyle="-",color=colors[i*3+j],marker="o")
        axs.legend(bbox_to_anchor=(1, 1), loc='upper left', borderaxespad=0.)
        plt.savefig(fname)
        plt.close()
        return    

    def write2csv(self,ijob,results1,results2):
        csvf = os.path.join(ijob,"stress_tensor1.csv")
        with open(csvf,"w") as f:
            f.write("The stress tensor calculated by analytical and finite difference method. Each x related to one configuration, and x is offset of cell vector.\n")
            f.write("x,")
            for i in range(3):
                for j in range(3):
                    f.write(f"analytical_{i+1}{j+1},")
                    f.write(f"numerical_{i+1}{j+1},")
            f.write("\n")
            for i in range(len(results1["x"])):
                f.write(f"{results1['x'][i]},")
                for j in range(3):
                    for k in range(3):
                        f.write(f"{results1[f'{j+1}{k+1}']['analytical'][i]},")
                        f.write(f"{results1[f'{j+1}{k+1}']['numerical'][i]},")
                f.write("\n")

        # save results2 to csv
        csvf = os.path.join(ijob,"stress_tensor2.csv")
        with open(csvf,"w") as f:
            f.write("The stress tensor calculated by finite difference method. x is the offset of cell vector.\n")
            f.write("x,")
            for i in range(3):
                for j in range(3):
                    f.write(f"{i+1}{j+1},")
            f.write("\n")
            for i in range(len(results2["x"])):
                f.write(f"{results2['x'][i]},")
                for j in range(3):
                    for k in range(3):
                        f.write(f"{results2[f'{j+1}{k+1}'][i]},")
                f.write("\n")        


import logging
from typing import Optional
from mpi4py import MPI

_comm: Optional[MPI.Comm] = None
_rank: int = -1
_size: int = -1
_initialized: bool = False

def init_mpi(comm: MPI.Comm = MPI.COMM_WORLD) -> None:
    """init MPI, only run onetime"""
    global _comm, _rank, _size, _initialized
    if _initialized:
        logging.warning("MPI has been initialized!!!")
        return
    
    _comm = comm
    _rank = comm.Get_rank()
    _size = comm.Get_size()
    _initialized = True
    logging.debug(f"MPI initialed: rank={_rank}, size={_size}")

def get_mpi_rank() -> int:
    """Get the local rank id"""
    if not _initialized:
        raise RuntimeError("Please run init_mpi() firstly!!!")
    return _rank

def get_mpi_size() -> int:
    """Get the total rank number"""
    if not _initialized:
        raise RuntimeError("Please run init_mpi() firstly!!!")
    return _size

def get_mpi_comm() -> MPI.Comm:
    """Get MPI communicator"""
    if not _initialized:
        raise RuntimeError("Please run init_mpi() firstly!!!")
    return _comm

def is_rank0() -> bool:
    """Check if current rank is rank 0"""
    return get_mpi_rank() == 0

def finalize_mpi() -> None:
    """finalize MPI"""
    global _comm, _rank, _size, _initialized
    if _initialized:
        _comm = None
        _rank = -1
        _size = -1
        _initialized = False


import subprocess
import os

BASE_VERSION = "v0.4.61"

def _no_need_subversion():
    try:
        modified_files = subprocess.check_output(
            ["git", "diff-tree", "--no-commit-id", "--name-only", "-r", "HEAD"],
            cwd=os.path.dirname(os.path.abspath(__file__)),
            stderr=subprocess.DEVNULL
        ).decode().splitlines()
        
        current_file = os.path.relpath(__file__, os.getcwd())
        
        # If this file is modified, indicating the main version is updated
        if current_file in modified_files:
            return True
        
        # If codes in the abacustest directory are modified, indicating the main version is not updated
        for i in modified_files:
            if i.startswith("abacustest/"):
                return False
        
        return True
    except:
        return False

def get_version():
    try:
        if _no_need_subversion():
            return BASE_VERSION
        
        commit_hash = subprocess.check_output(
            ["git", "rev-parse", "--short", "HEAD"],
            stderr=subprocess.STDOUT,
            text=True
        ).strip()
        

        timestamp = subprocess.check_output(
            ["git", "log", "-1", "--format=%cd", "--date=format:%Y%m%d%H%M%S"],
            stderr=subprocess.STDOUT,
            text=True
        ).strip()
        
        return f"{BASE_VERSION}+{timestamp}.{commit_hash}"
    except:
        return BASE_VERSION  

__version__ = get_version()


from dp.launching.typing.basic import BaseModel, Int, String, Float, List, Optional, Union, Dict
from dp.launching.typing import (
    BaseModel,
    Set,
    Boolean,
    Field,
    DflowAccessToken,
    DflowArgoAPIServer,
    DflowK8sAPIServer,
    DflowStorageEndpoint,
    DflowStorageRepository,
    BohriumMachineType,
    BohriumImage,
    BohriumPlatform,
    BohriumJobType,
    BohriumUsername,
    BohriumPassword,
    BohriumProjectId,
    BenchmarkLabels,
    BenchmarkTags
)
from enum import Enum
from typing import Literal
import re
from . import comm_class
from abacustest.constant import RECOMMAND_IMAGE, RECOMMAND_COMMAND, RECOMMAND_MACHINE


class Image(BaseModel):
    type: Literal["Local Image"]
    image: String = Field(default=RECOMMAND_IMAGE,
                          title="",
                          description="",)


class ImageBohrium(BaseModel):
    type: Literal["Use Bohrium Image"]

    image: String = Field(default=BohriumMachineType(RECOMMAND_IMAGE),
                          title="Bohrium Image Address",
                          description="",)
    bohrium_machine_type: BohriumMachineType = Field(default=BohriumMachineType(RECOMMAND_MACHINE))
    bohrium_job_type: BohriumJobType = Field(default=BohriumJobType.CONTAINER)
    bohrium_plat_form: BohriumPlatform = Field(default=BohriumPlatform.ALI)
    on_demand: Boolean = Field(default=False)
    #bohrium_machine_type: String = Field(default="c32_m128_cpu")
    #bohrium_job_type: String = Field(default="container")
    #bohrium_plat_form: String = Field(default="ali")


class RundftImageSet(BaseModel):
    rundft_image_set: Union[ImageBohrium,
                            comm_class.ImageDispatcher,
                     Image] = Field(discriminator="type",
                                    description="IMAGE and MACHINE used in RUN DFT")

class RundftGroupSizeSet(BaseModel):
    rundft_group_size: Int = Field(
        default=1, description="Number of examples running in each parallel machine. If not set, all examples will be run in parallel.", ge=0)

class RundftCommandSet(BaseModel):
    rundft_command: String = Field(default=RECOMMAND_COMMAND,
                                   description="Command to run each example. Please note that the program will first enter each folder before executing this command. \
During runtime, there will be no interaction between different examples",)

def parse_image_set(image_set):
    if isinstance(image_set, ImageBohrium):
        return {
            "image": image_set.image,
            "bohrium": {
                "scass_type": image_set.bohrium_machine_type,
                "job_type": image_set.bohrium_job_type,
                "platform": image_set.bohrium_plat_form,
                "on_demand": 1 if image_set.on_demand else 0
            }
        }
    elif isinstance(image_set, Image):
        return {
            "image": image_set.image
        }
    elif isinstance(image_set, comm_class.ImageDispatcher):
        return comm_class.ImageDispatcher.construct_dispatcher(image_set)
    else:
        raise ValueError(f"Unknown image set type: {type(image_set)}")

def construct_input(datas,opts,logs):
    # datas is a dict of examples, created by comm_class_exampleSource.read_source
    #read rundft
    need_rundft = False
    logs.iprint("read run dft setting ...")
    run_dft = [{}]
    
    #read rundft example
    if datas.get("rundft_example"):
        run_dft[-1]["example"] = datas.get("rundft_example")    
        logs.iprint("\texample:",run_dft[-1]["example"])
    
    #read rundft extra files
    if datas.get("rundft_extrafile"):
        run_dft[-1]["extra_files"] = datas.get("rundft_extrafile")
        logs.iprint("\rundft_extrafile:",run_dft[-1]["extra_files"])
        
    #read rundft command
    if hasattr(opts,"rundft_command"):
        need_rundft = True
        logs.iprint("\tcommand:",opts.rundft_command)
        run_dft[-1]["command"] = opts.rundft_command
    
    #read rundft ngroup        
    if hasattr(opts,"rundft_group_size") and opts.rundft_group_size > 0:
        logs.iprint("\tgroup size:",opts.rundft_group_size)
        run_dft[-1]["group_size"] = opts.rundft_group_size
    
    #read rundft image
    if hasattr(opts,"rundft_image_set"):
        image_set = parse_image_set(opts.rundft_image_set)   
        logs.iprint("\timage:",image_set.get("image",""))
        for k,v in image_set.items():
            run_dft[-1][k] = v
        if "bohrium" in run_dft[-1]:   
            logs.iprint("\tbohrium:",run_dft[-1]["bohrium"])
        elif "dispatcher" in run_dft[-1]:
            logs.iprint("\tdispatcher:",run_dft[-1]["dispatcher"]["resources_dict"])

    return need_rundft, run_dft


import sys,os,glob
import traceback
import importlib

def printAllMethod(allmethod,fmt):
    ipath = os.path.split(os.path.abspath(__file__))[0]
    print("Job type: %s, all modules: " % fmt,[os.path.split(i)[1][:-3] for i in glob.glob(os.path.join(ipath,"%s/*.py"%fmt))])
    print("\nmetric_name,\tmodule:function,\tdescription")
    for k,v in allmethod.items():
        filename = v[1] 
        funcname = v[0]
        out = filename.split("/")[-1][:-3] + ":" + funcname 
        print("%20s:\t%-30s\t%s"% (k,out,v[2]))
    print("\nNOTICE1: keys start with \"delta_\" require a json file includes the reference values. You can specify the file by \"--ref\"")
    print("NOTICE2: by default, only keys in module '%s' are available. You can use other modules by '--modules'" % fmt)

def import_new_method(newmethods=[]):
    '''import the self-defined methods'''
    if newmethods == None:
        return
    sys.path.append(os.getcwd())
    for imethod in newmethods:
        try:
            importlib.import_module(imethod)
            print("Import module: %s" % imethod)
        except:
            traceback.print_exc()
            print("Import module %s failed, skip it!" % imethod)
    #importlib.invalidate_caches()

def import_modules(fmt,modules):
    '''
    Can only import the modules in lib_collectdata.{fmt}
    '''
    if modules == None:
        return
    for module in modules:
        try:
            importlib.import_module("abacustest.lib_collectdata.%s.%s"%(fmt,module))
            print("import module abacustest.lib_collectdata.%s.%s" % (fmt,module))
        except:
            traceback.print_exc()
            print("Import module %s failed, skip it!" % module)

def RESULT(fmt="abacus",outparam=False,newmethods=[],modules=[],**kwargs):

    if fmt == "abacus":
        from .resultAbacus import ResultAbacus
        from .abacus import abacus
        import_modules(fmt,modules)
        import_new_method(newmethods)
            
        if outparam:
            printAllMethod(ResultAbacus.AllMethod(),fmt)
        else:
            return ResultAbacus(**kwargs)

    elif fmt == "qe":
        from .resultQe import ResultQe
        from .qe import qe
        import_modules(fmt,modules)
        import_new_method(newmethods)

        if outparam:
            printAllMethod(ResultQe.AllMethod(),fmt)
        else:
            return ResultQe(**kwargs)

    elif fmt == 'vasp':
        from .resultVasp import ResultVasp
        from .vasp import vasp
        import_modules(fmt,modules)
        import_new_method(newmethods)

        if outparam:
            printAllMethod(ResultVasp.AllMethod(),fmt)
        else:
            return ResultVasp(**kwargs)
    else:
        print("ERROR: unkonw format %s" % fmt)
        sys.exit(1)



import os,shutil,copy,argparse,sys,json,time

def gen_special_value():
    sv = {}
    date = time.strftime("%Y%m%d", time.localtime())
    sv["DATE"] = date
    return sv

def prepare_files(settings):
    # prepare the files, and then push to remote like github
    # 1. create the folder remote
    # 2. copy the files/folders to the folder
    # 3. support some special format:
    #   - {DATE}: the date of today, like 20210801
    '''
    "remote":{
        "files":{
            "file1_local": "file1_remote",
            "file2_local": "file2_remote",
            "file3_local": "/bda/{DATE}/fiel3_remote"
        }
    }
    '''
    if not settings:
        return None
    
    remote_folder = "remote"
    n = 1
    while os.path.isdir(remote_folder):
        remote_folder += str(n)
        n += 1
    os.mkdir(remote_folder)
    
    files_setting = settings.get("files",{})
    SV = gen_special_value()
    for ifile,jfile in files_setting.items():
        jfile = jfile.format(**SV)
        if not os.path.exists(ifile):
            print(f"Error: {ifile} does not exist!")
            continue
        if os.path.isfile(ifile):
            jdir = os.path.join(remote_folder,os.path.dirname(jfile))
            if not os.path.isdir(jdir):
                os.makedirs(jdir)
            shutil.copy(ifile,os.path.join(remote_folder,jfile))
        elif os.path.isdir(ifile):
            shutil.copytree(ifile,os.path.join(remote_folder,jfile),dirs_exist_ok=True)
    return remote_folder

def push_to_github(remote_path,github_setting,config_setting,extra_setting={}):
    # github_setting = 
    '''
        "github":{
            "repo": "",
            "branch": "",
            "commit_msg": ""
        }
        "add_date": "bda/dates"
    '''
    repo = github_setting.get("repo")
    branch = github_setting.get("branch")
    username = config_setting.get("github_username")
    email = config_setting.get("github_email")
    token = config_setting.get("github_token")
    commit_msg = github_setting.get("commit_msg","update from abacustest")
    
    if not repo:
        print("Error: the repo is not specified!")
        return 1
    if not branch:
        print("Error: the branch is not specified!")
        return 1
    if not username and not email and not token:
        print("Error: the username/email/token is not specified in config!")
        return 1

    remote_path = os.path.abspath(remote_path)
    
    github_path = "github"
    n = 1
    while os.path.isdir(github_path):
        github_path = "github" + str(n)
        n += 1
    os.mkdir(github_path)
    
    # add date to dates
    add_date_path = extra_setting.get("add_date")
    add_date_command = ""
    if add_date_path:
        date = time.strftime("%Y%m%d", time.localtime())
        add_date_command = f"""
# check if dates file exists
if [ -f "{add_date_path}" ]; then
    # check if the date exists
    if ! grep -q "{date}" "{add_date_path}"; then
        echo "{date}" >> "{add_date_path}"
    fi
else
    echo "{date}" > "{add_date_path}"
fi
        """
    
    # gen bash script
    repo_url_with_token=f"https://oauth2:{token}@github.com/{repo}.git"    
    bash_script = f'''#!/bin/bash
wget --no-proxy https://dp-public.oss-cn-beijing.aliyuncs.com/k8s/speedUp.sh -O speedUp.sh && bash speedUp.sh >> start.log
cd {github_path}
git init
git remote add origin "{repo_url_with_token}"

# Configure git user
git config user.name "{username}"
git config user.email "{email}"

# configure git debug mode
set GIT_TRACE=true
set GIT_CURL_VERBOSE=true
set GIT_SSH_COMMAND=ssh -vvv

# Checkout the repository
# try 10 times
fetch=0
for i in {{1..10}}
do
    git fetch --depth 1 origin {branch}
    if [ $? -eq 0 ]; then
        fetch=1
        break
    fi
    sleep 10
done
if [ $fetch -eq 0 ]; then
    echo "Error: can not fetch the branch {branch}!"
    cd ..
    exit 1
fi
git checkout {branch}

# copy the files
cp -r {remote_path}/* .

# check if add date
{add_date_command}

# commit and push
git add .
git commit -m "{commit_msg}"

# try 10 times
push=0
for i in {{1..10}}
do

    git push origin {branch}
    if [ $? -eq 0 ]; then
        push=1
        break
    fi  
    sleep 10
done
if [ $push -eq 0 ]; then
    echo "Error: can not push the branch {branch}!"
    cd ..
    exit 1
fi
cd ..
'''
    pushf = "push.sh"
    n=1
    while os.path.isfile(pushf):
        pushf = f"push{n}.sh"
        n+=1
    open(pushf,"w").write(bash_script)
    os.system(f"bash {pushf}")
    shutil.rmtree(github_path)
    os.remove(pushf)
    return 0

def push_to_remote(remote_path,setting,config_setting): 
    if setting.get("github"):
        push_to_github(remote_path,setting.get("github"),config_setting,setting)
    else:
        print("Error: the remote setting is not supported!")
        return 1

def Remote(param):
    param_file = param.param
    if not os.path.exists(param_file):
        print(f"Error: {param_file} does not exist!")
        sys.exit(1)
    settings = json.load(open(param_file))
    config_setting = settings.get("config",{})
    remote_setting = settings.get("remote",{})

    if remote_setting:
        remote_path = prepare_files(remote_setting)
        if remote_path:
            push_to_remote(remote_path,remote_setting,config_setting) 

def main():
    from abacustest.arguments import RemoteArgs
    parser = argparse.ArgumentParser()
    param = RemoteArgs(parser).parse_args()
    Remote(param)
    
if __name__ == "__main__":
    main()


#!/usr/bin/env python3
"""
Script to generate test cases for abacus collectdata.

Usage:
    python generate_test.py <path_to_abacus_example> [-o output_file.py]

Rules for test generation:
1. INPUT: Only verify it's a dict, and check calculation/esolver_type/kspacing values
2. Lists with >3 elements: Only verify length and first/middle/last elements
3. band_plot: Skip (path-dependent)
"""

import sys
import os
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent))

from abacustest.lib_collectdata.collectdata import RESULT

INPUT_SPECIAL_KEYS = ["calculation", "esolver_type", "kspacing"]
SKIP_METRICS = ["band_plot"]


def is_path_dependent(val):
    """Check if a value is a path that depends on cwd"""
    if not isinstance(val, str):
        return False
    path_indicators = ['/', '\\', '.png', '.txt', '.dat', '.xlsx', '.cube', '.html']
    return any(ind in val for ind in path_indicators) or Path(val).is_absolute()


def is_long_list(val):
    """Check if value is a list with more than 3 elements OR is deeply nested"""
    if not isinstance(val, list):
        return False
    if len(val) > 3:
        return True
    if len(val) > 0 and isinstance(val[0], list):
        depth = get_list_depth(val)
        if depth >= 2:
            return True
    return False


def generate_assertion(name, val):
    """Generate assertion line for a metric"""
    if val is None:
        return f"        # {name}: None"

    if name == "INPUT":
        return generate_input_assertion(val)
    elif name in SKIP_METRICS:
        return f"        # {name}: skipped"
    elif is_long_list(val):
        return generate_long_list_assertion(name, val)
    elif isinstance(val, bool):
        return f'        self.assertEqual(ab["{name}"], {val})'
    elif isinstance(val, int):
        return f'        self.assertEqual(ab["{name}"], {val})'
    elif isinstance(val, float):
        return f'        self.assertAlmostEqual(ab["{name}"], {val}, places=3)'
    elif isinstance(val, str):
        if is_path_dependent(val):
            return f'        # {name}: "{val}" (path-dependent, skipped)'
        return f'        self.assertEqual(ab["{name}"], "{val}")'
    elif isinstance(val, list) and len(val) == 0:
        return f'        self.assertEqual(ab["{name}"], [])'
    elif isinstance(val, list):
        if is_long_list(val):
            return generate_long_list_assertion(name, val)
        else:
            first = val[0]
            if isinstance(first, (int, float)):
                return f'        self.assertEqual(ab["{name}"], {val})'
            elif isinstance(first, str):
                return f'        self.assertEqual(ab["{name}"], {val})'
            else:
                return f'        self.assertEqual(ab["{name}"], {val})'
    elif isinstance(val, dict):
        return f'        self.assertEqual(ab["{name}"], {val})'
    else:
        return f'        # {name}: {type(val).__name__} = {val}'


def generate_input_assertion(val):
    """Generate assertion for INPUT metric"""
    lines = []
    lines.append("        input_dict = ab[\"INPUT\"]")
    lines.append("        self.assertIsInstance(input_dict, dict)")
    for key in INPUT_SPECIAL_KEYS:
        if key in val:
            v = val[key]
            if isinstance(v, str):
                lines.append(f'        self.assertEqual(input_dict["{key}"], "{v}")')
            else:
                lines.append(f'        self.assertEqual(input_dict["{key}"], {v})')
    return "\n".join(lines)


def get_list_depth(val):
    """Get nesting depth of a list"""
    if not isinstance(val, list):
        return 0
    if len(val) == 0:
        return 1
    return 1 + get_list_depth(val[0])


def flatten_list(val):
    """Flatten a nested list into 1D"""
    if not isinstance(val, list):
        return [val]
    result = []
    for item in val:
        result.extend(flatten_list(item))
    return result


def generate_long_list_assertion(name, val):
    """Generate assertion for long lists (>3 elements)"""
    length = len(val)
    lines = []

    first = val[0]
    depth = get_list_depth(val)

    if depth == 1:
        length = len(val)
        mid_index = length // 2
        middle = val[mid_index]
        lines.append(f"        self.assertEqual(len(ab[\"{name}\"]), {length})")
        if isinstance(first, (int, float)):
            lines.append(f"        self.assertAlmostEqual(ab[\"{name}\"][0], {first}, places=3)")
            lines.append(f"        self.assertAlmostEqual(ab[\"{name}\"][len(ab[\"{name}\"]) // 2], {middle}, places=3)")
            lines.append(f"        self.assertAlmostEqual(ab[\"{name}\"][-1], {val[-1]}, places=3)")
        elif isinstance(first, str):
            lines.append(f'        self.assertEqual(ab[\"{name}\"][0], "{first}")')
            lines.append(f'        self.assertEqual(ab[\"{name}\"][len(ab[\"{name}\"]) // 2], "{middle}")')
            lines.append(f'        self.assertEqual(ab[\"{name}\"][-1], "{val[-1]}")')
        else:
            lines.append(f"        # ab[\"{name}\"][0] = {first}")
    elif depth == 2:
        lines.append(f"        self.assertEqual(len(ab[\"{name}\"]), {length})")
        lines.append(f"        self.assertEqual(len(ab[\"{name}\"][0]), {len(first)})")
        flat = [item for sublist in val for item in sublist]
        lines.append(f"        {name}_flat = [item for sublist in ab[\"{name}\"] for item in sublist]")
        lines.append(f"        self.assertEqual(len({name}_flat), {len(flat)})")
        lines.append(f"        self.assertAlmostEqual({name}_flat[0], {flat[0]}, places=3)")
        lines.append(f"        self.assertAlmostEqual({name}_flat[len({name}_flat) // 2], {flat[len(flat) // 2]}, places=3)")
        lines.append(f"        self.assertAlmostEqual({name}_flat[-1], {flat[-1]}, places=3)")
    elif depth >= 3:
        lines.append(f"        self.assertEqual(len(ab[\"{name}\"]), {length})")
        lines.append(f"        self.assertEqual(len(ab[\"{name}\"][0]), {len(first)})")
        lines.append(f"        self.assertEqual(len(ab[\"{name}\"][0][0]), {len(first[0])})")
        flat = [item for sublist in val for subsublist in sublist for item in subsublist]
        lines.append(f"        {name}_flat = [item for sublist in ab[\"{name}\"] for subsublist in sublist for item in subsublist]")
        lines.append(f"        self.assertEqual(len({name}_flat), {len(flat)})")
        lines.append(f"        self.assertAlmostEqual({name}_flat[0], {flat[0]}, places=3)")
        lines.append(f"        self.assertAlmostEqual({name}_flat[len({name}_flat) // 2], {flat[len(flat) // 2]}, places=3)")
        lines.append(f"        self.assertAlmostEqual({name}_flat[-1], {flat[-1]}, places=3)")
    else:
        lines.append(f"        # ab[\"{name}\"][0] = {first}")

    return "\n".join(lines)


def generate_test(path, output_file=None):
    """Generate test code from abacus example"""
    ab = RESULT(path=path, fmt="abacus")

    from abacustest.lib_collectdata.abacus.abacus import Abacus
    all_names = sorted(Abacus.AllMethod().keys())

    not_none = []
    is_none = []

    for name in all_names:
        try:
            val = ab[name]
            if val is not None:
                not_none.append((name, val))
            else:
                is_none.append(name)
        except Exception:
            is_none.append(name)

    lines = []
    lines.append("import unittest")
    lines.append("from pathlib import Path")
    lines.append("")
    lines.append("from abacustest.lib_collectdata.collectdata import RESULT")
    lines.append("")
    lines.append("")
    lines.append("class TestAbacusCollectdata(unittest.TestCase):")
    lines.append('    """Test Abacus result collection from SCF calculation"""')
    lines.append("")
    lines.append("    def test_abacus_scf(self):")
    lines.append('        """Test all extractable metrics from abacus-scf example"""')
    lines.append(f'        ab = RESULT(path=Path(__file__).parent / "abacus-scf", fmt="abacus")')
    lines.append("")

    for name, val in not_none:
        lines.append(generate_assertion(name, val))

    lines.append("")
    lines.append("        # Metrics that returned None for this example:")
    for name in is_none:
        lines.append(f"        # {name}: None")

    lines.append("")
    lines.append("")
    lines.append("if __name__ == '__main__':")
    lines.append("    unittest.main()")

    result = "\n".join(lines)

    if output_file:
        with open(output_file, 'w') as f:
            f.write(result)
        print(f"Test file written to: {output_file}")
    else:
        print(result)

    return not_none, is_none


def main():
    if len(sys.argv) < 2:
        print("Usage: python generate_test.py <path_to_abacus_example> [-o output_file.py]")
        print("Example: python generate_test.py ./abacus-scf")
        print("Example: python generate_test.py ./abacus-scf -o test_output.py")
        sys.exit(1)

    path = sys.argv[1]
    output_file = None

    if len(sys.argv) >= 4 and sys.argv[2] == "-o":
        output_file = sys.argv[3]

    generate_test(path, output_file)


if __name__ == "__main__":
    main()



from .model import Model
import os, glob, json
from . import comm
import numpy as np
from abacustest.outresult import pandas_out
from abacustest.report import gen_html
from abacustest.lib_report.table import output_float
from .comm_plot import auto_set_yaxis


class CompareMetricsModel(Model):
    @staticmethod
    def add_args(parser):
        parser.add_argument("-f","--file",type=str,help="two metrics json files",action="extend",nargs=2,)
        parser.add_argument("-m","--metric",default=[],type=str,help="the metrics to be compared",action="extend",nargs="*",)
        parser.add_argument("-c","--const",default = [], type=str,help="the metrics of first metrics file will be shown. Usally be some comm metrics.",action="extend",nargs="*",)
        parser.add_argument("--show", default=0,const=1,help="show the example and metrics",nargs="?",)
        parser.add_argument("--name",type=str,help="The name of these two object, which is used in plot",default=None,action="extend",nargs=2,)
        parser.add_argument("--sort",default=1,const=1,help="If sort the example",nargs="?",)

        
    def run(self,params):
        m1f,m2f = params.file
        compare_metrics = params.metric
        const_metrics = params.const
        cm = CompareMetrics(m1f,m2f, compare_metrics, const_metrics,params.show, params.name,params.sort)
        cm.run()
        
                    
class CompareMetrics:
    def __init__(self, m1f,m2f, compare_metrics, const_metrics,show_m,name, sort):
        self.m1f = m1f
        self.m2f = m2f
        self.compare_metrics = compare_metrics
        self.const_metrics = const_metrics
        self.show_metrics = show_m
        self.name = name
        self.sort = sort
        print(f"Compare {m1f} and {m2f}")
        print(f"Metrics to compare: {compare_metrics}")
        print(f"Metrics to show: {const_metrics}")
    
    def run(self):
        if self.show_metrics:
            return self.show()
        
        if not os.path.exists(self.m1f):
            print(f"{self.m1f} does not exist.")
            return
        if not os.path.exists(self.m2f):
            print(f"{self.m2f} does not exist.")
            return

        m1 = json.load(open(self.m1f))
        m2 = json.load(open(self.m2f))
        self.m1 = m1
        self.m2 = m2
        
        report, sm = self.cal_dev(m1,m2)
        self.plot_metrics(report,"compare_metrics.png")
        json.dump({"fig_results":{"type": "image", "file": "compare_metrics.png"}}, open("supermetrics.json", "w"), indent=4)
        
        json.dump(report,open("report.json","w"),indent=4)    
        json.dump(sm,open("summary.json","w"),indent=4)
        print( "The report is saved in report.json")
        print(f"The results of {self.m1f} and {self.m2f} are compared, and the deviation between them.") 
        print( "  Dev means the difference between '{self.m1f}' and '{self.m2f}'.")
        print( "  MaxDev means the maximum absolue difference between two values.")
        print(f"  Ratio means '{self.m1f}' devided by '{self.m2f}'")
        print("")
        pandas_out(report) 
        print("")
        print("Summary:")
        print(sm)
        
        # gen html
        gen_html({"content": [{"type": "supermetrics", "content": "summary.json"},
                              {"type": "metrics","content": "report.json"}]
                  },
                 "abacustest.html")
        print("The html file is saved in abacustest.html")   
    
    def is_float(self, v):
        try:
            float(v)
            return True
        except:
            return False
    
    def plot_metrics(self,report,fname):
        nmetrics = len(self.compare_metrics)
        if nmetrics == 0:
            print("No metrics to plot")
            return

        examples = list(report.keys())
        if self.sort:
            examples.sort()
        
        if self.name is None:
            label1 = self.m1f
            label2 = self.m2f
        else:
            label1 = self.name[0]
            label2 = self.name[1]
            
        # now, we need to calculate how many figs we need
        plot_values = []
        for im in self.compare_metrics:
            x1 = range(len(examples))
            x2 = range(len(examples))
            
            y1 = [self.m1[ie].get(im,None) for ie in examples]
            y2 = [self.m2[ie].get(im,None) for ie in examples]
            x1, y1 = comm.clean_none_list(x1,y1)
            x2, y2 = comm.clean_none_list(x2,y2)
            if len(x1) == 0 or len(x2) == 0:
                continue
            # check if y is str
            #if any([isinstance(iv,str) for iv in y1+y2]):
            #    continue
            
            # only plot the scalar
            if any([not self.is_float(iv) for iv in y1+y2]):
                continue
            
            plot_values.append({"name":im,"x1": x1, "y1": y1, "x2": x2, "y2": y2})
            if f"Dev({im})" in report[examples[0]]:
                dev_x = range(len(examples))
                dev_y = [report[ie][f"Dev({im})"] for ie in examples]
                dev_x, dev_y = comm.clean_none_list(dev_x,dev_y)
                plot_values[-1]["sub_x"] = dev_x
                plot_values[-1]["sub_y"] = dev_y
                plot_values[-1]["sub_name"] = f"Dev({im})"
            if f"Ratio({im})" in report[examples[0]]:
                dev_ratio_x = range(len(examples))
                dev_ratio_y = [report[ie][f"Ratio({im})"] for ie in examples]
                dev_ratio_x, dev_ratio_y = comm.clean_none_list(dev_ratio_x,dev_ratio_y)
                plot_values.append({"name":im, "x1": x1, "y1": y1, "x2": x2, "y2": y2, 
                                        "sub_x": dev_ratio_x, "sub_y": dev_ratio_y, "sub_name": f"Ratio({im})"})
            
        ncol = min(2, len(plot_values))
        nrow = len(plot_values)//ncol
        while nrow*ncol < len(plot_values): nrow += 1   
        
        width = max(6, 0.4*len(examples))
        if width > 12:
            ncol = 1
            nrow = len(plot_values)
        
        import matplotlib.pyplot as plt
        
        fig, axs = plt.subplots(nrow,ncol,figsize=(width*ncol,6*nrow))
        fontsize = 18
        for i,plot_value in enumerate(plot_values):
            metric = plot_values[i]["name"]
            if nrow == 1 and ncol == 1:
                ax = axs
            elif nrow == 1 or ncol == 1:
                ax = axs[i]
            else:
                ax = axs[i//ncol,i%ncol]
            ax.plot(plot_value["x1"],plot_value["y1"],"r-o",label=label1)
            ax.plot(plot_value["x2"],plot_value["y2"],"b-^",label=label2)
            ax.legend(loc="upper left",fontsize=fontsize)
            title = metric
            if "sub_name" in plot_value:
                title += f", {plot_value['sub_name']}"
                if plot_value['sub_name'].startswith("Dev"):
                    title += f" (MeanDev: {output_float(self.cal_mean(plot_value['sub_y']))})"
                elif plot_value['sub_name'].startswith("Ratio"):
                    title += f" (GeoMeanRatio: {output_float(self.cal_mean(plot_value['sub_y'],geom=True))})"
            ax.set_title(title,fontsize=fontsize+2)
            ax.set_xlabel("Example",fontsize=fontsize+2)
            ax.set_ylabel(metric,fontsize=fontsize+2)
            ax.set_xticks(range(len(examples)))
            ax.set_xticklabels(examples,rotation=30,fontsize=fontsize)
            ax.tick_params(axis='y',labelsize=fontsize)
            auto_set_yaxis(ax, plot_value["y1"]+plot_value["y2"], margin1=0.1, margin2=0.3)
            
            if "sub_x" in plot_value:
                ax2 = ax.twinx()
                ax2.plot(plot_value["sub_x"],plot_value["sub_y"],"g--",label=plot_value["sub_name"],marker='o',markerfacecolor='none')
                ax2.set_ylabel(plot_value["sub_name"],fontsize=fontsize+2)
                ax2.legend(loc="upper right",fontsize=fontsize)
                ax2.tick_params(axis='y',labelsize=fontsize)
                auto_set_yaxis(ax2, plot_value["sub_y"], margin1=0.1, margin2=0.3)
                
            
        plt.tight_layout()
        plt.savefig(fname,dpi=400)
        plt.close()

    def cal_mean(self, values, geom=False):
        values = comm.clean_none_list(values)
        if len(values) == 0:
            return None
        if geom:
            return np.exp(np.mean(np.log(values)))
        else:
            return np.mean(values)

    def show(self):
        for mf in [self.m1f,self.m2f]:
            print(f"File: {mf}")
            m = json.load(open(mf))
            examples = list(m.keys())
            examples.sort()
            examples = " ".join(examples)
            print(f"Examples: {examples}")
            metrics = []
            for ie in m.keys():
                metrics += m[ie].keys()
            metrics = list(set(metrics))
            metrics.sort()
            metrics = "\" \"".join(metrics)
            print(f"Metrics: \"{metrics}\"")

    def cal_dev(self, m1, m2):
        examples = list(set(m1.keys()) & set(m2.keys()))
        if self.sort:
            examples.sort()
        
        e1 = [[m1.get(im,{}).get(ik, None) for im in examples] for ik in self.compare_metrics]
        e2 = [[m2.get(im,{}).get(ik, None) for im in examples] for ik in self.compare_metrics]
        dev = [[] for i in range(len(self.compare_metrics))]
        ratio = [[] for i in range(len(self.compare_metrics))]
        for i in range(len(self.compare_metrics)):
            for j in range(len(examples)):
                v1 = e1[i][j]
                v2 = e2[i][j]
                idev, idevratio = self.cal_one_dev(v1,v2)
                dev[i].append(idev)
                ratio[i].append(idevratio)
        
        # check if the dev or ratio is None
        has_dev = [any([iv is not None for iv in idev]) for idev in dev]
        has_dev_ratio = [any([iv is not None for iv in idev]) for idev in ratio]
        
        report = {} 
        for idx_e, ie in enumerate(examples):
            report[ie] = {}
            for im in self.const_metrics:
                report[ie][im] = m1.get(ie,{}).get(im,None)
            for idx_m, im in enumerate(self.compare_metrics):
                report[ie][f"{im}_1"] = m1.get(ie,{}).get(im,None)
                report[ie][f"{im}_2"] = m2.get(ie,{}).get(im,None)
                if has_dev[idx_m]:
                    report[ie][f"Dev({im})"] = dev[idx_m][idx_e]
                if has_dev_ratio[idx_m]:
                    report[ie][f"Ratio({im})"] = ratio[idx_m][idx_e]
        
        sm = {}
        for i, im in enumerate(self.compare_metrics):
            if has_dev[i]:
                dev_nonone = [iv for iv in dev[i] if iv is not None]
                if len(dev_nonone) > 0:
                    sm[f"MeanDev({im})"] = self.cal_mean(dev_nonone)
                    sm[f"MaxDev({im})"] = max(dev_nonone)
                    sm[f"MinDev({im})"] = min(dev_nonone)
            if has_dev_ratio[i]:
                ratio_nonone = [iv for iv in ratio[i] if iv is not None]
                if len(ratio_nonone) > 0:
                    sm[f"GeoMeanRatio({im})"] = self.cal_mean(ratio_nonone,geom=True)
                    sm[f"MaxRatio({im})"] = max(ratio_nonone)
                    sm[f"MinRatio({im})"] = min(ratio_nonone)
                                          
        return report, sm             

    def cal_one_dev(self,v1,v2):
        if None in [v1,v2]:
            return None,None
        try:
            if isinstance(v1,str) or isinstance(v2,str):
                return f"{v1}|{v2}",None
            elif isinstance(v1,(int,float)) and isinstance(v2,(int,float)):
                if v2 == 0:
                    return v1-v2, None
                else:
                    return v1-v2, v1/v2
            else:
                return abs(np.array(v1)-np.array(v2)).max().tolist(), None
        except:
            return None, None          
    

    


from .abacus import AbacusStru, ReadInput, ReadKpt
import os
from . import comm,cp2k
import copy

def update_dict(dict1, dict2):
    '''
    Update dict1 with dict2, if the key in dict1 is also in dict2, then update the value of dict1 with the value of dict2.
    If the value of one key in dict1 and dict2 are both dict, then merge the two dicts.
    
    Need check the case for section&value, such as KIND&element, which should be treated as a section of KIND.
    For one key not in dict1:
        If the key is type section&value, then
            If key is in repeat_section, then we add the section to dict1.
            else, check if key is in dict1, if yes, del the key in dict1, and add the new section to dict1.
        else
            add the key to dict1.
    
    For repeat_section,
    '''
    if (not isinstance(dict1,dict)) or (not isinstance(dict2,dict)):
        return dict2
    repeat_section = ["KIND"]
    for k,v in dict2.items():
        if k not in dict1:
            if "&" in k:
                real_k = k.split("&")[0]
                if real_k in repeat_section and v != None:
                    dict1[k] = v
                elif real_k in dict1:
                    if v != None:
                        dict1[k] = update_dict(dict1[real_k],v)
                    del dict1[real_k]
                else:
                    dict1_keys = [i.split("&")[0] for i in dict1.keys()]
                    if real_k not in dict1_keys:
                        if v != None:
                            dict1[k] = v
                    else:
                        idx = dict1_keys.index(real_k)
                        del dict1[list(dict1.keys())[idx]]
                        if v!= None:
                            dict1[k] = v
            else: 
                dict1_keys = [i.split("&")[0] for i in dict1.keys()]
                if k not in dict1_keys:
                    if v != None:  
                        dict1[k] = v
                else:
                    idx = dict1_keys.index(k)
                    if v != None:
                        dict1[k] = update_dict(dict1[list(dict1.keys())[idx]],v)
                    del dict1[list(dict1.keys())[idx]]
        else:
            if isinstance(v,dict) and isinstance(dict1[k],dict):
                dict1[k] = update_dict(dict1[k],v)
            elif v == None:
                del dict1[k]
            else:
                dict1[k] = v
    return dict1

def param2cp2k(param:dict, save_path:str=None):
    '''
    write the param dict to cp2k input file.
    
    The param dict is a nested dict, the key is the section name, the value is the parameters in this section.
    For each key and value:
        1. If the value is a dict, it is a section, and add "&" before the section name, and "&END" after the section.
            Special for the "&KIND" section, the key should be "KIND&element", and the value should be a dict, which contains the element, basis_set, potential,....
        2. If the value is a list, it is a list of values.
        3. If the value is a string or int or float, it is a parameter.
    '''
    def write_section(iparam, layer=0):
        cc = ""
        for k,v in iparam.items():
            if isinstance(v,dict):
                if "&" in k:
                    ks = k.split("&")
                    cc += "  "*layer + f"&{ks[0]} {ks[1]}\n"
                    cc += write_section(v,layer+1)
                    cc += "  "*layer + f"&END {ks[0]}\n"
                else:
                    cc += "  "*layer + f"&{k}\n"
                    cc += write_section(v,layer+1)
                    cc += "  "*layer + f"&END {k}\n"
                if layer < 2:
                    cc += "\n"
            elif isinstance(v,list):
                cc += "  "*layer + f"{k} "+ " ".join([str(i) for i in v]) + "\n"
            elif isinstance(v,(str,int,float)):
                cc += "  "*layer + f"{k} {v}\n"
            else:
                print(f"Warning: the value of {k} is {v}, which type ({type(v)}) is not supported now, just transfer {v} to a string.")
                cc += "  "*layer + f"{k} {v}\n"
        return cc
    cc = write_section(param)
    if save_path:
        print(f"Write CP2K input to {save_path}")
        with open(save_path,"w") as f:
            f.write(cc)
    return cc

def kpt2cp2k(kpt,kpt_model):
    if kpt_model in ["gamma","mp"]:
        return {"FORCE_EVAL": {"DFT": {"KPOINTS": {"SCHEME": "MONKHORST-PACK " + " ".join([str(i) for i in kpt[:3]])}}}}
    elif kpt_model == "direct":
        return {"FORCE_EVAL": {"DFT": {"KPOINTS": {"KPOINT": kpt}}}}
    else:
        print("Warning: do not support KPT model {model} now! Do not set the kpoints.")
        return {}
    

def stru2cp2k(stru:AbacusStru):
    '''
    Convert abacus structure to cp2k input.
    
    Args:
        stru: the abacus structure
    '''
    cell = stru.get_cell(bohr=False)
    coord = stru.get_coord(bohr=False,direct=False)
    label = stru.get_label(total=True)
    element = stru.get_element(number=False,total=True)
    mag = stru.get_atommag()
    
    tmp = {
        "FORCE_EVAL": {
            "SUBSYS": {
                "CELL": {
                    "A": cell[0],
                    "B": cell[1],
                    "C": cell[2],
                },
                "COORD": {
                    label[i] : coord[i] for i in range(len(label))
                },
            },
            "DFT": {
                "BASIS_SET_FILE_NAME": "BASIS_MOLOPT",
                "POTENTIAL_FILE_NAME": "GTH_POTENTIALS",
            }
        }
    }
    
    label_uniq = []
    for idx,ilabel in enumerate(label):
        if ilabel not in label_uniq:
            label_uniq.append(ilabel)
            tmp["FORCE_EVAL"]["SUBSYS"][f"KIND&{ilabel}"] = {
                "ELEMENT": element[idx],
                "MAGNETIZATION": mag[idx],
                "BASIS_SET": "DZVP-MOLOPT-GTH" if element[idx] not in cp2k.BASIS else cp2k.BASIS[element[idx]],
                "POTENTIAL": "GTH-BLYP" if  element[idx] not in cp2k.POTENTIAL else cp2k.POTENTIAL[element[idx]],
            }
    return tmp

def calculation2cp2k(calculation, force, stress):
    if calculation in ["scf","md","relax","cell-relax"]:
        tmp = {"GLOBAL": {"RUN_TYPE": {"scf": "ENERGY", "md": "MD", "relax": "GEO_OPT","cell-relax":"CELL_OPT"}[calculation]}}
    else:
        print(f"Warning: the value of calculation is {calculation}, which set GLOBAL/RUN_TYPE to ENERGY")
        tmp = {"GLOBAL": {"RUN_TYPE": "ENERGY"}}
    
    if calculation == "scf" and force:
        tmp["GLOBAL"]["RUN_TYPE"] = "ENERGY_FORCE"
    
    if stress or calculation == "cell-relax":
        tmp["FORCE_EVAL"] = {"STRESS_TENSOR": "ANALYTICAL"}
    
    print_section = {}
    if calculation in ["relax", "cell-realx"] or force:
        print_section["FORCES&ON"] = {}
    if calculation in ["cell-relax"] or stress:
        print_section["STRESS_TENSOR&ON"] = {}
    if print_section:
        if "FORCE_EVAL" not in tmp:
            tmp["FORCE_EVAL"] = {}
        tmp["FORCE_EVAL"]["PRINT"] = print_section
    return tmp

def esolver2cp2k(esolver,dft_functional):
    tmp = {}
    if esolver == "ksdft":
        tmp = {"FORCE_EVAL": {"METHOD": "QS"}}
    else:
        print(f"Warning: the value of esolver is {esolver}, which is not supported now. Set FORCE_EVAL/METHOD to QS.")
        tmp = {"FORCE_EVAL": {"METHOD": "QS"}}
    
    dft_func = dft_functional.upper()
    tmp["FORCE_EVAL"]["DFT"] = {"XC": {f"XC_FUNCTIONAL&{dft_func}": {}}}
    
    return tmp

def kssolver2cp2k(kssolver):
    if kssolver == None:
        return {}
    tmp = {}
    if kssolver.lower() in ["genelpa","scalapack_gvx","cg", "dav"]:
        tmp = {"FORCE_EVAL": {"DFT": {"SCF": {"DIAGONALIZATION":{"ALGORITHM": "STANDARD"}}}}}
    else:
        print(f"Warning: the value of ks_solver is {kssolver}, which is not supported now. Set FORCE_EVAL/DFT/SCF/ALGORITHM to STANDARD.")
        tmp = {"FORCE_EVAL": {"DFT": {"SCF": {"DIAGONALIZATION":{"ALGORITHM": "STANDARD"}}}}}
    return tmp

def mix2cp2k(mixing_type,mixing_beta, mixing_ndim):
    if mixing_type == None:
        return {}
    
    tmp = {"FORCE_EVAL": {"DFT": {"SCF": {"MIXING": {}}}}}
    if mixing_type == "plain":
        tmp["FORCE_EVAL"]["DFT"]["SCF"]["MIXING"]["METHOD"] = "DIRECT_P_MIXING"
    elif mixing_type == "pulay":
        tmp["FORCE_EVAL"]["DFT"]["SCF"]["MIXING"]["METHOD"] = "PULAY_MIXING"
    elif mixing_type == "broyden":
        tmp["FORCE_EVAL"]["DFT"]["SCF"]["MIXING"]["METHOD"] = "BROYDEN_MIXING"
    else:
        print(f"Warning: the value of mixing_type is {mixing_type}, which is not supported now. Set FORCE_EVAL/DFT/SCF/MIXING/METHOD to DIRECT_P_MIXING.")
    
    if mixing_beta != None:
        tmp["FORCE_EVAL"]["DFT"]["SCF"]["MIXING"]["ALPHA"] = mixing_beta
    
    if mixing_ndim != None:
        tmp["FORCE_EVAL"]["DFT"]["SCF"]["MIXING"]["NBUFFER"] = mixing_ndim
    
    return tmp

def smearing2cp2k(smearing_method, smearing_sigma):
    if smearing_method == None:
        return {}
    
    if smearing_method == "fd":
        tmp = {"FORCE_EVAL": {"DFT": {"SCF": {"SMEAR&T": {"METHOD": "FERMI_DIRAC"}}}}}
    else:
        print(f"Warning: the value of smearing_method is {smearing_method}, which is not supported now. Will not set smear.")
        return {}
    
    return tmp

def scf2cp2k(scf_thr, scf_nmax, basis): 
    '''
    As the scf_thr in ABACUS is for difference of charge, while in CP2K is for energy, so based on a general sense,
    if basis is PW, then set the EPS_SCF to scf_thr*1e3,
    else, basis is lcao, then set the EPS_SCF to scf_thr*1e2
    '''
    if scf_thr == None:
        return {}
    
    if basis == None or basis == "pw":
        tmp = {"FORCE_EVAL": {"DFT": {"SCF": {"EPS_SCF": scf_thr*1e3}}}}   
    elif basis in ["lcao","lcao_in_pw"]:
        tmp = {"FORCE_EVAL": {"DFT": {"SCF": {"EPS_SCF": scf_thr*1e2}}}}
    else: 
        print(f"Warning: the value of basis is {basis}, which is not supported now. Will not set EPS_SCF.")
        return {}
    if scf_nmax != None:
        tmp["FORCE_EVAL"]["DFT"]["SCF"]["MAX_SCF"] = scf_nmax
    return tmp

def relaxthr2cp2k(calculation,force_thr, force_thr_ev,stress_thr):
    # MAX_FORCE bohr^-1*hartree
    if calculation not in ["relax","cell-relax"]:
        return {}
    
    if calculation == "relax":
        sub_section = "GEO_OPT"
    else:
        sub_section = "CELL_OPT"
        
    force = None
    if force_thr_ev != None:
        force = force_thr_ev * 0.03889 / 2
    if force_thr:
        force = force_thr / 2
    
    thr_set = {}
    if force != None:
        thr_set["MAX_FORCE"] = force
        thr_set["RMS_FORCE"] = force*2/3
    if stress_thr != None:
        thr_set["PRESSURE_TOLERANCE"] = stress_thr*1000
    
    if thr_set:
        return {"MOTION": {sub_section: thr_set}}
    else:
        return {}
    
def plusu2cp2k(plusu=None):
    return {}

def vdw2cp2k(vdw=None):
    if vdw == None:
        return {}   
            
def spin2cp2k(nspin):
    if nspin == None or nspin == 1:
        return {}
    elif nspin ==2:
        return {"FORCE_EVAL": {"DFT": {"UKS": "T"}}}
    else:
        print(f"Warning: the value of nspin is {nspin}, which is not supported now. Will not set UKS.")
        return {}

def input2cp2k(input_param:dict):
    
    # one by one parameter dict:
    oneone = {
        "suffix": lambda x: {"GLOBAL": {"PROJECT": x}},
    }
    
    tmp = {"GLOBAL": {"PRINT_LEVEL": "MEDIUM"}}
    
    # transfer oneone
    allkeys = list(input_param.keys())
    for k in allkeys:
        k = k.lower()
        if k in oneone:
            itmp = oneone[k](input_param[k])
            if itmp is not None:
                update_dict(tmp,itmp)
            del input_param[k]
    
    # transfer calculation, force, stress
    calculation = input_param.pop("calculation","scf")
    update_dict(tmp,calculation2cp2k(calculation,
                                        input_param.pop("cal_force",False),
                                        input_param.pop("cal_stress",False)))
    update_dict(tmp,scf2cp2k(input_param.pop("scf_thr",None), input_param.pop("scf_nmax",None), input_param.pop("basis_type",None)))
    update_dict(tmp,smearing2cp2k(input_param.pop("smearing_method",None), input_param.pop("smearing_sigma",None)))
    update_dict(tmp,mix2cp2k(input_param.pop("mixing_type",None), input_param.pop("mixing_beta",None), input_param.pop("mixing_ndim",None)))
    update_dict(tmp,kssolver2cp2k(input_param.pop("ks_solver",None)))
    update_dict(tmp,esolver2cp2k(input_param.pop("esolver","ksdft"), input_param.pop("dft_functional","pbe")))
    update_dict(tmp,plusu2cp2k(input_param.pop("plusu",None)))
    update_dict(tmp,vdw2cp2k(input_param.pop("vdw",None)))
    update_dict(tmp,spin2cp2k(input_param.pop("nspin",None)))
    update_dict(tmp,relaxthr2cp2k(calculation,input_param.pop("force_thr",None), input_param.pop("force_thr_ev",None), input_param.pop("stress_thr",None)))
    
    not_supportted_keys = list(input_param.keys())
    if len(not_supportted_keys) > 0:
        print(f"Warning: below keys are not supported now, and will ignore them:\n    ", ", ".join(not_supportted_keys))
    
    return tmp
    
def Abacus2Cp2k(abacus_path:str, save_path:str=None, cp2k_setting={}):
    '''
    Convert abacus input to cp2k input.
    
    1. read the abacus input file, and STRU/KPT
    2. generate the cp2k input file
    
    Args:
        abacus_path: the path of the abacus input file
        save_path: the path to save the cp2k input file, default is the same as the abacus input file
        cp2k_setting: the extra setting for cp2k input file
    '''
    cp2k_setting = copy.deepcopy(cp2k_setting)
    tmp = {}
    abacus_input = ReadInput(os.path.join(abacus_path,"INPUT"))
    abacus_stru = AbacusStru.ReadStru(os.path.join(abacus_path,"STRU"))
    abacus_kpt, kpoint_model = ReadKpt(os.path.join(abacus_path,"KPT"))
    
    if abacus_input == None or abacus_stru == None:
        print("Error: the abacus input file or stru file is not found.")
        return
    
    update_dict(tmp,stru2cp2k(abacus_stru))
    update_dict(tmp,kpt2cp2k(abacus_kpt,kpoint_model))
    update_dict(tmp,input2cp2k(abacus_input))
    update_dict(tmp,cp2k_setting)
    
    if save_path == None:
        save_path = os.path.join(abacus_path,"cp2k.inp")
    
    param2cp2k(tmp,save_path)
    
    
    


from .model import Model
import json, os, shutil
from pathlib import Path
from typing import List, Dict, Any, Literal, Optional, Tuple

import numpy as np

from abacustest import AbacusSTRU, WriteInput, ReadInput, RESULT
from abacustest.constant import (
    RECOMMAND_IMAGE,
    RECOMMAND_COMMAND,
    RECOMMAND_MACHINE,
    RECOMMAND_VASP_IMAGE,
    RECOMMAND_VASP_COMMAND,
    RECOMMAND_VASP_MACHINE,
)
from abacustest.lib_model.comm import copy_abacusjob, get_largest_vacuum_dir
from abacustest.lib_data.grid import Potential
from abacustest.lib_prepare.abacus2vasp import Abacus2Vasp
from abacustest.lib_collectdata.collectdata import RESULT as CollectDataRESULT

EFIELD_DIRECTION_MAP = ["a", "b", "c"]


class WorkFuncModel(Model):
    @staticmethod
    def prepare_args(parser):
        """
        Add arguments for the prepare subcommand
        The arguments can not be command, model, modelcommand
        """
        parser.description = "Prepare the inputs for the work function calculation."
        parser.add_argument(
            "-j",
            "--job",
            default=[],
            action="extend",
            nargs="*",
            help="the paths of ABACUS jobs, should contain INPUT, STRU, or KPT, and pseudopotential and orbital files",
        )
        parser.add_argument(
            "--vacuum",
            default="auto",
            choices=["a", "b", "c", "auto"],
            help="The direction of the vacuum for work function calculation, default is auto. If set to auto, the direction will try to be determined from structure automatically.",
        )
        parser.add_argument(
            "--dipole-corr",
            action="store_true",
            help="Whether to use dipole correction for the work function calculation, default is not using.",
        )
        parser.add_argument(
            "--image",
            type=str,
            default=None,
            help="The image to use for the Bohrium job. Default is none, which means determine by used DFT software."
        )
        parser.add_argument(
            "--machine",
            type=str,
            default=None,
            help="The machine to use for the Bohrium job. Default is none, which means determine by used DFT software.",
        )
        parser.add_argument(
            "--dft-command",
            type=str,
            default=None,
            help=f"The command to run the dft job. Default is none, which means determine by used DFT software.",
        )
        parser.add_argument(
            "--dft",
            default="abacus",
            choices=["abacus", "vasp"],
            help="DFT software to use for work function calculation (ABACUS or VASP). Default is ABACUS.",
        )
        parser.add_argument(
            "--potcar_path",
            type=str,
            default=None,
            help="Path to VASP POTCAR files directory (required for VASP input generation)",
        )
        parser.add_argument(
            "--use-empty-atom",
            action="store_true",
            help="Use empty atom in vacuum region near surface at both sides of the surface. Default is False. Only works for ABACUS.",
        )
        parser.add_argument(
            "--empty-atom-elem",
            type=str,
            default=None,
            help="The element of the empty atom. Default is None, which means use first kind of element in the structure. Only works for ABACUS and when --use-empty-atom is set to True.",
        )
        parser.add_argument(
            "--empty-atom-height",
            type=float,
            default=2.0,
            help="The height of the empty atom layer in Angstrom. Default is 2.0 Angstrom. Only works for ABACUS and when --use-empty-atom is set to True.",
        )
        parser.add_argument(
            "--empty-atom-dist",
            type=float,
            default=2.0,
            help="The distance of empty atom along directions along lattice vectors parallel to the surface. Default is 2.0 Angstrom. Only works for ABACUS and when --use-empty-atom is set to True.",
        )
        return parser

    def run_prepare(self, params):
        """
        Run the model with the given parameters
        """
        if not params.job:
            raise ValueError("No job specified, please use -j or --job to specify the job paths.")
        
        paths = prep_all_workfunc_jobs(params.job, params.dft, params.vacuum, params.dipole_corr, params.potcar_path,
                                       params.use_empty_atom, params.empty_atom_elem, params.empty_atom_height, params.empty_atom_dist)
        
        if params.image is None:
            image = RECOMMAND_IMAGE if params.dft == "abacus" else RECOMMAND_VASP_IMAGE if params.dft == "vasp" else None
        else:
            image = params.image
        if params.machine is None:
            machine = RECOMMAND_MACHINE if params.dft == "abacus" else RECOMMAND_VASP_MACHINE if params.dft == "vasp" else None
        else:
            machine = params.machine
        if params.dft_command is None:
            dft_command = RECOMMAND_COMMAND if params.dft == "abacus" else RECOMMAND_VASP_COMMAND if params.dft == "vasp" else None
        else:
            dft_command = params.dft_command

        # Create setting file
        if paths:
            setting = {"save_path": "results", "run_dft": []}
            setting["run_dft"].append(
                {
                    "ifrun": True,
                    "example": paths,
                    "command": dft_command,
                    "image": image,
                    "bohrium": {
                        "scass_type": machine,
                        "job_type": "container",
                        "platform": "ali",
                    },
                }
            )

        setting_file = "setting.json"
        json.dump(setting, open(setting_file, "w"), indent=4)
        print(f"\nInputs are generated in: {', '.join([str(p) for p in paths])}")
        print(f"You can modify '{setting_file}', and execute below command to run the abacustest to submit all jobs to bohrium:\n\tabacustest submit -p {setting_file}\n")

        # Print postprocess instructions
        print(f"\nAfter finishing the calculations, you can postprocess the work function results:")
        if params.dft == "abacus":
            print(f"  For ABACUS results: abacustest model workfunc post -j {' '.join(params.job)} --dft abacus")
        elif params.dft == "vasp":
            print(f"  For VASP results: abacustest model workfunc post -j {' '.join(params.job)} --dft vasp")

    @staticmethod
    def postprocess_args(parser):
        """
        Add arguments for the postprocess subcommand
        The arguments can not be command, model, modelcommand
        """
        parser.description = "Postprocess the work function calculation results."
        parser.add_argument(
            "-j",
            "--job",
            default=[],
            action="extend",
            nargs="*",
            help="the paths of the job directories, should contain the results of work function calculations generated by the prepare step.",
        )
        parser.add_argument(
            "--vacuum",
            default="auto",
            choices=["a", "b", "c", "auto"],
            help="The direction of the vacuum for work function calculation, default is auto. If set to auto, the direction will try to be determined from input file and structure automatically.",
        )
        parser.add_argument(
            "--dft",
            default="abacus",
            choices=["abacus", "vasp"],
            help="DFT software used for work function calculation (ABACUS or VASP)",
        )
        parser.add_argument(
            "--thr",
            default=0.01,
            type=float,
            help="Threshold for derivative of averaged potential for identifying potential plateaus in work function calculation.",
        )
        return parser

    def run_postprocess(self, params):
        """
        Parse the parameters and run the postprocess process"""
        if not params.job:
            raise ValueError(
                "No job specified, please use -j or --job to specify the job paths."
            )

        print("Work function results (units in eV):")

        results_all = {}
        for job in params.job:
            if not os.path.isdir(job):
                raise ValueError("The job path is not a directory: %s" % job)

            # Determine job type based on parameters
            jobtype = "vasp" if params.dft == "vasp" else "abacus"
            results, plot_path, pot_file, plot_data_file = post_workfunc_calc(
                job, jobtype=jobtype, vacuum_dir_specified=params.vacuum
            )

            work_functions = ""
            for result in results:
                work_functions += f"{result['work_function']:4f} eV, "
            print(f"{job}: {work_functions}")

            results_all[job] = results

        json.dump(results_all, open("metrics_workfunc.json", "w"), indent=4)
        print("\nThe postprocess is done. The metrics are saved in 'metrics.json'.")


def prep_all_workfunc_jobs(jobs: List[str], 
                           mode: Literal['abacus', 'vasp'], 
                           vacuum_dir: Literal['a', 'b', 'c', 'auto'] = 'auto', 
                           dipole_corr: bool = False, 
                           potcar_path: Optional[str] = None,
                           use_empty_atom: bool = False,
                           empty_atom_elem: Optional[str] = None,
                           empty_atom_height: float = 2.0,
                           empty_atom_dist: float = 2.0) -> List[str]:
    paths = []
    for job in jobs:
        if not os.path.isdir(job):
            raise ValueError("The job path is not a directory: %s" % job)

        # Prepare ABACUS calculation if ABACUS is selected
        if mode == "abacus":
            abacus_path = prep_abacus_workfunc_calc(job, vacuum_dir, dipole_corr, os.path.join(job, "workfunc_job"),
                                                    use_empty_atom, empty_atom_elem, empty_atom_height, empty_atom_dist)
            paths.append(abacus_path)

        # Prepare VASP calculation if VASP is selected
        elif mode == "vasp":
            vasp_path = prep_vasp_workfunc_calc(job, vacuum_dir, dipole_corr, potcar_path)
            paths.append(vasp_path)
    
    return paths
        

def prep_abacus_workfunc_calc(
    job: Path,
    vacuum_dir: Literal["a", "b", "c", "auto"] = "auto",
    dipole_corr: bool = False,
    workfunc_dir: Optional[str] = "workfunc_job",
    use_empty_atom: bool = False,
    empty_atom_elem: Optional[str] = None,
    empty_atom_height: float = 2.0,
    empty_atom_dist: float = 2.0,
) -> Path:
    workfunc_dir = Path(workfunc_dir).absolute()
    if workfunc_dir.exists():
        print("Old workfunc_job directory found, remove and recreate it.")
        shutil.rmtree(Path(job) / "workfunc_job")

    copy_abacusjob(job, workfunc_dir, out_dir=False)

    if use_empty_atom:
        add_empty_atom_layer(workfunc_dir, elem=empty_atom_elem, height=empty_atom_height, dist=empty_atom_dist)

    input_params = ReadInput(os.path.join(workfunc_dir, "INPUT"))
    if input_params.get("nspin", 1) not in [1, 2]:
        raise ValueError("Only 1 or 2 spin is supported.")

    input_params["calculation"] = "scf"
    input_params["out_pot"] = 2

    if dipole_corr:
        input_params["efield_flag"] = 1
        input_params["dip_cor_flag"] = 1
        input_params['efield_amp'] = 0.0 # Automatically determine electric field strength
        input_params['efield_pos_max'] = None # Override original electric field position
        input_params['efield_pos_dec'] = None # Override original range to decrease of electric field
        
        # Identify vacuum direction
        stru_file = os.path.join(workfunc_dir, input_params.get("stru_file", "STRU"))
        stru = AbacusSTRU.read(stru_file)
        direction, max_vacuum_cart, vacuum_top, vacuum_bottom = get_largest_vacuum_dir(coords=stru.coords, cell=stru.cell)

        # Identify vacuum direction to apply electric field
        if vacuum_dir == "auto":
            print(f"Using the largest vacuum direction identified by the largest vacuum: %s" % direction)
            input_params["efield_dir"] = EFIELD_DIRECTION_MAP.index(direction)
        elif vacuum_dir in ["a", "b", "c"]:
            if direction != vacuum_dir:
                print(f"WARNING: The given vacuum direction is {vacuum_dir}, not the same with the one identified by the largest vacuum {direction}.")
            print(f"Using the provided vacuum direction {vacuum_dir}")
            input_params["efield_dir"] = EFIELD_DIRECTION_MAP.index(vacuum_dir) # Always use the provided vacuum direction
        else:
            raise ValueError("Invalid vacuum direction: %s" % vacuum_dir)
    
    WriteInput(input_params, os.path.join(workfunc_dir, "INPUT"))

    return str(workfunc_dir)


def prep_vasp_workfunc_calc(
    job: Path,
    vacuum_dir: Literal["a", "b", "c", "auto"] = "auto",
    dipole_corr: bool = False,
    potcar_path: Optional[str] = None,
) -> Path:
    """
    Prepare VASP work function calculation from ABACUS input.

    Args:
        job: Path to ABACUS job directory
        vacuum_dir: Vacuum direction for work function calculation
        dipole_corr: Whether to use dipole correction
        potcar_path: Path to VASP POTCAR files directory

    Returns:
        Path to prepared VASP work function calculation directory
    """
    import uuid

    # Check and remove existing VASP directory
    vasp_dir_name = "workfunc_vasp_job"
    if vasp_dir_name in os.listdir(job):
        print(f"Old {vasp_dir_name} directory found, remove and recreate it.")
        shutil.rmtree(Path(job) / vasp_dir_name)

    # Check and remove existing ABACUS directory if it exists
    abacus_dir_name = "workfunc_job" + str(uuid.uuid4())[:8] # avoid confilct with existed ABACUS job
    if abacus_dir_name in os.listdir(job):
        print(f"Existing {abacus_dir_name} directory found, removing it.")
        shutil.rmtree(Path(job) / abacus_dir_name)

    # First create ABACUS workfunc job to get proper INPUT parameters
    abacus_workfunc_dir = prep_abacus_workfunc_calc(job, vacuum_dir, dipole_corr, abacus_dir_name)

    # Create VASP directory
    vasp_workfunc_dir = Path(os.path.join(job, vasp_dir_name)).absolute()
    os.makedirs(vasp_workfunc_dir, exist_ok=True)

    # Convert ABACUS input to VASP input
    try:
        if potcar_path is None:
            potcar_path = os.environ.get("VASP_PP_PATH", None)
        if potcar_path is None:
            print("Warning: No POTCAR provided")
        Abacus2Vasp(
            abacus_path=str(abacus_workfunc_dir),
            save_path=str(vasp_workfunc_dir),
            potcar=potcar_path,
            vasp_setting={"emax_coef": 1.5},
        )
    except Exception as e:
        raise RuntimeError(f"Failed to convert ABACUS to VASP input: {e}")

    # Remove the temporary ABACUS directory
    shutil.rmtree(abacus_workfunc_dir)

    return str(vasp_workfunc_dir)

def add_empty_atom_layer(
    job: Path,
    elem: Optional[str] = None,
    height: float = 2.0,
    dist: float = 2.0,
) -> None:
    """
    Add empty atom in vacuum region near surface at both sides of the surface, and write the modified structure to the job directory.
    """
    from abacustest.lib_prepare.stru import AbacusATOM

    input_params = ReadInput(os.path.join(job, "INPUT"))
    stru_file = os.path.join(job, input_params.get("stru_file", "STRU"))
    stru = AbacusSTRU.read(stru_file)

    if elem is None:
        elem = stru.atomtypes[0].element
        pp, orb = stru.atomtypes[0].pp, stru.atomtypes[0].orb
    elif elem in stru.elements:
        # Get pp, orb from structure
        for atomtype in stru.atomtypes:
            if atomtype.element == elem:
                pp, orb = atomtype.pp, atomtype.orb
                break
    else:
        raise ValueError("Empty atom not in the original structure not supported yet")
    
    vac_dir, max_vacuum_cart, vacuum_top, vacuum_bottom = get_largest_vacuum_dir(coords=stru.coords, cell=stru.cell)
    # Assume the vacuum direction is perpendicular to surface
    a1, a2, a3 = np.array(stru.cell[0]), np.array(stru.cell[1]), np.array(stru.cell[2])
    # sa1, sa2 are lattice vectors parallel to surface
    # va is the lattice vector perpendicular to surface
    if vac_dir == "a":
        sa1, sa2, va = a2, a3, a1
    elif vac_dir == "b":
        sa1, sa2, va = a1, a3, a2
    elif vac_dir == "c":
        sa1, sa2, va = a1, a2, a3
    else:
        raise ValueError("Invalid vacuum direction: %s" % vac_dir)
    
    l1, l2 = np.linalg.norm(sa1), np.linalg.norm(sa2)
    # center of the surface containing empty atoms
    top_empty_layer_vec = va / np.linalg.norm(va) * (vacuum_top - height)
    bottom_empty_layer_vec = va / np.linalg.norm(va) * (vacuum_bottom + height)
    top_center_pos = top_empty_layer_vec + sa1/2 + sa2/2
    bottom_center_pos = bottom_empty_layer_vec + sa1/2 + sa2/2

    # Calculate number of atoms along 2 direction parallel to surface
    # Distance between atoms in different cell are assured to be larger than dist
    n1, n2 = int(l1/dist), int(l2/dist)
    
    # Displacement vector relative to middle point of the lattice vector parallel to surface
    d1s = [((-n1+1)/2 + i) * sa1/l1*dist for i in range(n1)]
    d2s = [((-n2+1)/2 + i) * sa2/l2*dist for i in range(n2)]

    empty_atom_coords_top = [top_center_pos + d1 + d2 for d1 in d1s for d2 in d2s]
    empty_atom_coords_bottom = [bottom_center_pos + d1 + d2 for d1 in d1s for d2 in d2s]

    for coord in empty_atom_coords_top:
        stru._atoms.append(AbacusATOM(label=elem+"_empty", coord=coord, pp=pp, orb=orb))
    for coord in empty_atom_coords_bottom:
        stru._atoms.append(AbacusATOM(label=elem+"_empty", coord=coord, pp=pp, orb=orb))
    
    stru.write(stru_file)

def post_workfunc_calc(
    job: Path, jobtype: Literal["abacus", "vasp"] = "abacus",
    vacuum_dir_specified: Literal["a", "b", "c", "auto"] = "auto",
    thr: float = 0.01,
) -> Tuple[List[Dict[str, Any]], Path, str, str]:
    # Postprocess calculation to obtain work function data
    if jobtype == "abacus":
        workfunc_job = os.path.join(job, "workfunc_job")
        if not os.path.exists(workfunc_job):
            workfunc_job = Path(job).absolute()
        results = RESULT(fmt="abacus", path=workfunc_job)
        efermi = results["efermi"]

        if results["normal_end"] is not True or results["converge"] is not True:
            raise RuntimeError(
                "ABACUS calculation didn't end normally or didn't reached SCF converge."
            )

        input_params = ReadInput(os.path.join(workfunc_job, "INPUT"))
        vacuum_dir_input = input_params.get("efield_dir", None)
        if vacuum_dir_input:
            vacuum_dir_input = EFIELD_DIRECTION_MAP[vacuum_dir_input]
        
        pot_file = os.path.join(
            workfunc_job,
            f"OUT.{input_params.get('suffix', 'ABACUS')}/ElecStaticPot.cube",
        )
        pot = Potential.from_cube(pot_file)

        stru_file = os.path.join(workfunc_job, input_params.get("stru_file", "STRU"))
        stru = AbacusSTRU.read(stru_file)
        vacuum_dir_find = get_largest_vacuum_dir(coords=stru.coords, cell=stru.cell)[0]

        vacuum_dir = select_vacuum_dir(vacuum_dir_input, vacuum_dir_find, vacuum_dir_specified)
        
        ave_elec_stat, coord_direct = pot.profile1d(axis=vacuum_dir, average=True)
        ave_elec_stat = -ave_elec_stat  # consider negative charge of electron
    elif jobtype == "vasp":
        from ase.io import read
        from pymatgen.io.vasp import Incar

        workfunc_job = os.path.join(job, "workfunc_vasp_job")

        # Use lib_collectdata to get VASP calculation results
        try:
            result = CollectDataRESULT("vasp", path=workfunc_job)
            # Accessing keys will automatically call the registered methods
            efermi = result["efermi"]

            if efermi is None:
                # Check if OUTCAR exists
                outcar_file = os.path.join(workfunc_job, "OUTCAR")
                if not os.path.exists(outcar_file):
                    raise RuntimeError(f"VASP OUTCAR file not found in {workfunc_job}. Please run the VASP calculation first.")
                else:
                    raise RuntimeError("Could not find Fermi energy in VASP results. Calculation may have failed.")
        except Exception as e:
            raise RuntimeError(f"Failed to collect VASP results: {str(e)}")
        # Read LOCPOT file using Potential.from_locpot
        locpot_file = os.path.join(workfunc_job, "LOCPOT")
        if not os.path.exists(locpot_file):
            raise RuntimeError(f"VASP LOCPOT file not found in {workfunc_job}")

        pot = Potential.from_locpot(locpot_file)

        pos = read(os.path.join(workfunc_job, "POSCAR"))
        vacuum_dir_find = get_largest_vacuum_dir(coords=pos.get_positions(), cell=pos.get_cell())[0]
        incar_params = Incar.from_file(os.path.join(workfunc_job, "INCAR"))
        idipol = incar_params.get("IDIPOL", None)
        if idipol is None:
            vacuum_dir_input = None
        else:
            vacuum_dir_input = EFIELD_DIRECTION_MAP[incar_params.get("IDIPOL", None) - 1]

        print(vacuum_dir_input, vacuum_dir_find, vacuum_dir_specified)
        vacuum_dir = select_vacuum_dir(vacuum_dir_input, vacuum_dir_find, vacuum_dir_specified)

        # Get averaged electrostatic potential profile
        ave_elec_stat, coord_direct = pot.profile1d(axis=vacuum_dir, average=True)
        ave_elec_stat = -ave_elec_stat  # consider negative charge of electron

        pot_file = locpot_file
    else:
        raise ValueError(f"Unsupported job type: {jobtype}")

    profile_data_file = os.path.join(workfunc_job, "profiled.dat")
    with open(profile_data_file, "w") as f:
        f.write(f"{'fraction_coord':>16s}{'ave_elec_stat':>16s}\n")
        for i in range(len(coord_direct)):
            f.write(f"{coord_direct[i]:16.8f}{ave_elec_stat[i]:16.8f}\n")

    # Get cell length along vacuum direction for proper slope calculation
    axis_idx = 0 if vacuum_dir == "a" else 1 if vacuum_dir == "b" else 2
    cell_length = np.linalg.norm(pot.cell[axis_idx])

    work_function_results = calculate_work_functions(ave_elec_stat, fermi_energy=efermi,
                                                     cell_length=cell_length, thr=thr)

    plot_path = plot_averaged_elecstat_pot(
        coord_direct,
        ave_elec_stat,
        efermi,
        axis=vacuum_dir,
        work_function_results=work_function_results,
        plot_filename=os.path.join(workfunc_job, "elecstat_pot_profile.png"),
        title=os.path.basename(job),
    )

    return work_function_results, plot_path, pot_file, profile_data_file



def identify_potential_plateaus(averaged_potential: List, cell_length: float, threshold: float = 0.01):
    """
    Identify plateaus in the electrostatic potential using derivative of the averaged electrostatic potential.

    Args:
        averaged_potential: The averaged electrostatic potential values in eV
        cell_length: The length of the cell along the vacuum direction in Angstrom
        threshold: Threshold for the slope (in eV/Angstrom) to be considered as plateau, default is 0.01
    """
    pot_derivatives = []
    n_points = len(averaged_potential)
    # Use cartesian step size (in Angstrom) instead of fractional coordinate
    stepsize = cell_length / (n_points - 1)
    for i in range(n_points):
        backward_idx = i - 1
        forward_idx = 0 if i == n_points - 1 else i + 1
        pot_derivative = (
            averaged_potential[forward_idx] - averaged_potential[backward_idx]
        ) / (stepsize * 2.0)
        pot_derivatives.append(pot_derivative)

    is_plateau = [abs(deriv) < threshold for deriv in pot_derivatives]

    plateau_ranges = []
    in_plateau, start_idx = False, None
    for i in range(n_points):
        if is_plateau[i] and not in_plateau:
            in_plateau = True
            start_idx = i
        elif not is_plateau[i] and in_plateau:
            in_plateau = False
            if not start_idx == i - 1:
                plateau_ranges.append((start_idx, i - 1))
                start_idx = None
        elif is_plateau[i] and i == n_points - 1:
            if not start_idx == i - 1:
                plateau_ranges.append((start_idx, i))
                in_plateau, start_idx = False, None

    if len(plateau_ranges) > 0:
        if plateau_ranges[-1][1] == n_points - 1:
            if len(plateau_ranges) > 1:
                combined_plateau = (
                    plateau_ranges[-1][0] - n_points,
                    plateau_ranges[0][1],
                )
                plateau_ranges[0] = combined_plateau
                plateau_ranges.pop()

    return plateau_ranges


def calculate_work_functions(averaged_potential: List, fermi_energy: float, cell_length: float, thr: float = 0.01):
    """
    Calculate the work function from the averaged electrostatic potential.
    Dipole correction is suppoted and multiple plateau of electrostatic potential can be identified.
    """
    work_function_results = []
    plateau_ranges = identify_potential_plateaus(averaged_potential, cell_length, threshold=thr)
    if len(plateau_ranges) == 0:
        # If no plateau is found, use the largest value of the averaged potential as the vacuum level to calculate work function, and give a warning message.
        print("Warning: No plateaus found in the averaged electrostatic potential. \nUse the maximum value of the averaged potential as the vacuum level to calculate work function, which may be less accurate.")

        # Find the maximum value and its index in the averaged potential
        averaged_potential_np = np.array(averaged_potential)
        max_potential_idx = np.argmax(averaged_potential_np)
        max_potential_value = averaged_potential_np[max_potential_idx]
        
        # Create a result using the maximum value
        npoints = len(averaged_potential)
        result = {
            "plateau_start_fractional": max_potential_idx / npoints,
            "plateau_end_fractional": (max_potential_idx + 1) / npoints,
            "vacuum_level": max_potential_value,
            "work_function": max_potential_value - fermi_energy,
        }
        work_function_results.append(result)
        
    else:
        npoints = len(averaged_potential)
        for plateau_range in plateau_ranges:
            plateau_start, plateau_end = plateau_range
            result = {
                "plateau_start_fractional": plateau_start / npoints,
                "plateau_end_fractional": plateau_end / npoints
            }
            if plateau_start < 0:
                plateau_start = plateau_start % npoints
            if plateau_end >= npoints:
                plateau_end  = plateau_end % npoints

            if plateau_start > plateau_end:
                #If plateau crosses the end of the cell, concatenate the plateau
                plateau = np.concatenate((averaged_potential[plateau_start:], averaged_potential[:plateau_end]))
            else:
                plateau = averaged_potential[plateau_start:plateau_end]

            plateau_averaged_potential = np.average(plateau)
            result['vacuum_level'] = plateau_averaged_potential
            result['work_function'] = plateau_averaged_potential - fermi_energy
            work_function_results.append(result)

    return work_function_results


def plot_averaged_elecstat_pot(
    coord: np.ndarray,
    averaged_elecstat_data: np.ndarray,
    efermi: float,
    axis: Literal["a", "b", "c"] = "c",
    work_function_results: Optional[List[Dict[str, float]]] = None,
    plot_filename: Optional[str] = "elecstat_pot_profile.png",
    title: Optional[str] = None,
) -> Path:
    import matplotlib.pyplot as plt

    plt.figure(figsize=(8, 4))
    plt.plot(coord, averaged_elecstat_data, label="Average electrostatic potential")
    plt.axhline(
        y=efermi,
        linestyle="--",
        color="gray",
        alpha=0.5,
        label=f"Fermi Energy (={efermi:.2f} eV)",
    )
    plt.xlim(min(coord), max(coord))
    plt.xlabel("Fractional coordinate along " + axis)
    plt.ylabel("Electrostatic potential (eV)")

    # Add title if provided
    if title:
        plt.title(title)

    if work_function_results is not None:
        for result in work_function_results:
            # Calculate position of center of plateau
            start_idx = int(result["plateau_start_fractional"] * (len(coord) - 1))
            end_idx = int(result["plateau_end_fractional"] * (len(coord) - 1))
            mid_idx = (start_idx + end_idx) // 2
            x_mid = coord[mid_idx]

            plateau_potential = averaged_elecstat_data[mid_idx]
            work_func = result["work_function"]

            # Draw arrow between plateau and Fermi energy
            plt.annotate(
                "",
                xy=(x_mid, efermi),
                xytext=(x_mid, plateau_potential),
                arrowprops=dict(arrowstyle="<->", color="black", lw=1.5),
            )

            # Decide position of annotated text
            x_range = max(coord) - min(coord)
            offset = x_range * 0.02
            if x_mid + offset <= max(coord):
                text_x = x_mid + offset
                ha_align = "left"
            else:
                text_x = x_mid - offset
                ha_align = "right"

            text_y = (efermi + plateau_potential) / 2

            plt.text(
                text_x,
                text_y,
                f"{work_func:.2f} eV",
                fontsize=10,
                ha=ha_align,
                va="center",
            )

    plt.legend(loc="best")
    plt.tight_layout()
    plt.savefig(plot_filename, dpi=300)
    plt.close()

    return Path(plot_filename).absolute()

def select_vacuum_dir(vacuum_dir_input, vacuum_dir_find, vacuum_dir_specified):
    """
    select used vacuum direction from different sources.
    """
    if vacuum_dir_specified == "auto":
        if vacuum_dir_input != vacuum_dir_find:
            print(f"WARNING: The given vacuum direction is {vacuum_dir_input}, not the same with the vacuum direction {vacuum_dir_find} identified from the structure.")
            if vacuum_dir_input is None:
                print(f"Using the direction {vacuum_dir_find} identified from structure.")
                vacuum_dir = vacuum_dir_find
            else:
                print(f"Using the direction {vacuum_dir_input} from input file.")
                vacuum_dir = vacuum_dir_input
        else:
            print(f"Using the direction {vacuum_dir_input} from input file.")
            vacuum_dir = vacuum_dir_input
    elif vacuum_dir_specified in ['a', 'b', 'c']:
        if not vacuum_dir_input == vacuum_dir_find == vacuum_dir_specified:
            print(f"WARNING: The given vacuum direction {vacuum_dir_specified}, direction from input file {vacuum_dir_input},  and identified vacuum from structure {vacuum_dir_find} are not all the same.")
            print(f"Using the direction {vacuum_dir_specified} from user input.")
            vacuum_dir = vacuum_dir_specified
        else:
            vacuum_dir = vacuum_dir_input
    else:
        raise ValueError("Invalid specified vacuum direction: %s" % vacuum_dir_specified)
    
    return vacuum_dir



import traceback
from dp.launching.typing.basic import BaseModel
import os

from dp.launching.report import Report,AutoReportElement,ReportSection,ChartReportElement


from . import (comm_class,
               comm_func,
               comm_class_exampleSource,
               comm_class_prepare,
               comm_class_predft,
               readsetting)


class PredftModel(
    comm_class_predft.PredftImageSet,
    comm_class_predft.PredftCommandSet,
    comm_class_exampleSource.PredftExtraFileSet,
    #comm_class_predft.PredftGroupSizeSet,
    comm_class_prepare.PrepareSet,
    comm_class_prepare.PrepareOrbLibSet,
    comm_class_prepare.PreparePPLibSet,
    comm_class_prepare.PrepareDPKSDescriptorSet,
    comm_class_prepare.PrepareKptTemplateSet,
    comm_class_prepare.PrepareStruTemplateSet,
    comm_class_prepare.PrepareInputTemplateSet,
    comm_class_exampleSource.PrepareExtraFileSet,
    comm_class_exampleSource.PrepareExampleSet,
    comm_class_exampleSource.PrepareExampleSourceSet,
    comm_class.OutputSet,
    comm_class.ConfigSet,
        BaseModel):
    ...  

class PredftDatasetsModel(
    comm_class_predft.PredftImageSet,
    comm_class_predft.PredftCommandSet,
    comm_class_exampleSource.PredftExtraFileNeededSet,
    #comm_class_predft.PredftGroupSizeSet,
    comm_class_prepare.PrepareSet,
    comm_class_prepare.PrepareOrbLibPathSet,
    comm_class_prepare.PreparePPLibPathSet,
    comm_class_prepare.PrepareDPKSDescriptorPathSet,
    comm_class_prepare.PrepareKptTemplatePathSet,
    comm_class_prepare.PrepareStruTemplatePathSet,
    comm_class_prepare.PrepareInputTemplatePathSet,
    comm_class_exampleSource.PrepareExtraFileNeededSet,
    comm_class_exampleSource.PrepareExampleSet,
    comm_class_exampleSource.DatasetSet,
    comm_class.OutputSet,
    comm_class.ConfigSet,
        BaseModel):
    ...  

def PredftModelRunner(opts) -> int:
    try:
        logs = comm_class.myLog()

        paths = comm_func.create_path(str(opts.IO_output_path))
        output_path = paths["output_path"]
        work_path = paths["work_path"]
        download_path = paths["download_path"]

        allparams = readsetting.ReadSetting(logs,opts,work_path,download_path)
        if allparams == None:
            return 1

        stdout,stderr = comm_func.exec_abacustest(allparams,work_path)
        logs.iprint(f"{stdout}\n{stderr}\nrun abacustest over!\n")
        
        #move results to output_path
        comm_func.move_results_to_output(work_path,output_path,allparams.get("save_path","results"))
        comm_func.pack_results(output_path,allparams.get("save_path","results"))
        
    except:
        traceback.print_exc()
        return 1

    return 0



import unittest
import numpy as np
import os
import shutil
import tempfile
from pathlib import Path

from abacustest.lib_prepare import abacus


class TestAbacusStru(unittest.TestCase):
    """Test AbacusStru class"""

    def setUp(self):
        self.work_dir = tempfile.mkdtemp()
        self.cell = [[2.87, 0, 0], [0, 2.87, 0], [0, 0, 2.87]]
        self.coord = [[0, 0, 0], [1.435, 1.435, 1.435]]
        self.stru = abacus.AbacusStru(
            label=["Fe"],
            cell=self.cell,
            coord=self.coord,
            atom_number=[2],
            pp=["Fe.upf"],
            element=["Fe"],
            magmom=[1.0],
            lattice_constant=1.0,
        )

    def tearDown(self):
        if os.path.exists(self.work_dir):
            shutil.rmtree(self.work_dir)

    def test_get_natoms(self):
        self.assertEqual(self.stru.get_natoms(), 2)

    def test_get_pp(self):
        pp = self.stru.get_pp()
        self.assertEqual(pp, ["Fe.upf"])

    def test_get_pp_total(self):
        pp_total = self.stru.get_pp(total=True)
        self.assertEqual(len(pp_total), 2)

    def test_get_orb(self):
        orb = self.stru.get_orb()
        self.assertIsNone(orb)

    def test_get_paw(self):
        paw = self.stru.get_paw()
        self.assertIsNone(paw)

    def test_get_label(self):
        label = self.stru.get_label(total=True)
        self.assertEqual(label, ["Fe", "Fe"])

    def test_get_label_not_total(self):
        label = self.stru.get_label(total=False)
        self.assertEqual(label, ["Fe"])

    def test_get_mass(self):
        mass = self.stru.get_mass()
        self.assertEqual(len(mass), 1)
        self.assertGreater(mass[0], 0)

    def test_get_element(self):
        element = self.stru.get_element(total=True, number=False)
        self.assertEqual(element, ["Fe", "Fe"])

    def test_get_element_with_number(self):
        element = self.stru.get_element(total=True, number=True)
        self.assertEqual(element[0], 26)

    def test_get_cell(self):
        cell = self.stru.get_cell(bohr=False)
        self.assertEqual(len(cell), 3)

    def test_get_cell_bohr(self):
        cell = self.stru.get_cell(bohr=True)
        self.assertEqual(len(cell), 3)

    def test_get_cell_param(self):
        a, b, c, alpha, beta, gamma = self.stru.get_cell_param()
        self.assertGreater(a, 0)
        self.assertAlmostEqual(alpha, 90.0)

    def test_get_volume(self):
        volume = self.stru.get_volume()
        self.assertGreater(volume, 0)

    def test_get_coord(self):
        coord = self.stru.get_coord(bohr=False, direct=False)
        self.assertEqual(len(coord), 2)

    def test_get_coord_direct(self):
        coord = self.stru.get_coord(bohr=False, direct=True)
        self.assertEqual(len(coord), 2)

    def test_get_mag(self):
        mag = self.stru.get_mag()
        self.assertEqual(mag, [1.0])

    def test_get_move(self):
        move = self.stru.get_move()
        self.assertEqual(len(move), 6)

    def test_globalidx2labelidx(self):
        label, idx = self.stru.globalidx2labelidx(0)
        self.assertEqual(label, "Fe")
        self.assertEqual(idx, 0)

    def test_globalidx2labelidx_second_atom(self):
        label, idx = self.stru.globalidx2labelidx(1)
        self.assertEqual(label, "Fe")
        self.assertEqual(idx, 1)

    def test_globalidx2labelidx_invalid_index(self):
        with self.assertRaises(ValueError):
            self.stru.globalidx2labelidx(10)

    def test_globalidx2labelidx_negative_index(self):
        with self.assertRaises(ValueError):
            self.stru.globalidx2labelidx(-1)

    def test_labelidx2globalidx(self):
        idx = self.stru.labelidx2globalidx("Fe", 1)
        self.assertEqual(idx, 1)

    def test_labelidx2globalidx_invalid_label(self):
        with self.assertRaises(AssertionError):
            self.stru.labelidx2globalidx("Invalid", 0)

    def test_set_pp(self):
        self.stru.set_pp(["NewFe.upf"])
        pp = self.stru.get_pp()
        self.assertEqual(pp, ["NewFe.upf"])

    def test_set_orb(self):
        self.stru.set_orb(["Fe.orb"])
        orb = self.stru.get_orb()
        self.assertEqual(orb, ["Fe.orb"])

    def test_set_mass(self):
        self.stru.set_mass([56.0])
        mass = self.stru.get_mass()
        self.assertEqual(mass[0], 56.0)

    def test_set_element(self):
        self.stru.set_element(["Co"])
        element = self.stru.get_element(total=False, number=True)
        self.assertEqual(element[0], 27)

    def test_set_atommag(self):
        self.stru.set_atommag([2.0, 2.0])
        mag = self.stru.get_atommag()
        self.assertEqual(mag, [2.0, 2.0])

    def test_set_atommag_none(self):
        self.stru.set_atommag(None)
        mag = self.stru.get_atommag()
        self.assertEqual(mag, [1.0, 1.0])

    def test_set_angle1(self):
        self.stru.set_angle1([90.0, 90.0])
        angle1 = self.stru.get_angle1()
        self.assertEqual(angle1, [90.0, 90.0])

    def test_set_angle2(self):
        self.stru.set_angle2([45.0, 45.0])
        angle2 = self.stru.get_angle2()
        self.assertEqual(angle2, [45.0, 45.0])

    def test_set_constrain(self):
        self.stru.set_constrain([[True, True, False], [False, False, False]])
        constrain = self.stru.get_constrain()
        self.assertEqual(constrain[0], [True, True, False])

    def test_get_constrain(self):
        constrain = self.stru.get_constrain()
        self.assertEqual(len(constrain), 2)

    def test_get_isconstrain(self):
        self.stru.set_constrain([True, False])
        is_constrain = self.stru.get_isconstrain()
        self.assertTrue(is_constrain[0])

    def test_get_lambda(self):
        self.stru.set_constrain([True, False])
        lambda_vals = self.stru.get_lambda()
        self.assertEqual(len(lambda_vals), 2)

    def test_get_atommag(self):
        mag = self.stru.get_atommag()
        self.assertEqual(len(mag), 2)

    def test_get_atommag_norm(self):
        mag = self.stru.get_atommag(norm=True)
        self.assertEqual(len(mag), 2)

    def test_set_dpks(self):
        self.stru.set_dpks("jle.orb")
        dpks = self.stru.get_dpks()
        self.assertEqual(dpks, "jle.orb")

    def test_supercell_2x1x1(self):
        new_stru = self.stru.supercell([2, 1, 1])
        self.assertEqual(new_stru.get_natoms(), 4)

    def test_supercell_2x2x2(self):
        new_stru = self.stru.supercell([2, 2, 2])
        self.assertEqual(new_stru.get_natoms(), 16)

    def test_delete_atom_all(self):
        new_stru = self.stru.delete_atom("Fe")
        self.assertEqual(new_stru.get_natoms(), 0)

    def test_delete_atom_by_index(self):
        new_stru = self.stru.delete_atom("Fe", idx=0)
        self.assertEqual(new_stru.get_natoms(), 1)

    def test_set_empty_atom(self):
        self.stru.set_empty_atom("Fe", idx=0)
        label = self.stru.get_label(total=False)
        self.assertIn("Fe_empty", label)

    def test_set_empty_atom_all(self):
        self.stru.set_empty_atom("Fe")
        label = self.stru.get_label(total=False)
        self.assertIn("Fe_empty", label)

    def test_write_stru(self):
        stru_file = os.path.join(self.work_dir, "test.STRU")
        self.stru.write(stru_file)
        self.assertTrue(os.path.exists(stru_file))

    def test_write_stru_direct(self):
        stru_file = os.path.join(self.work_dir, "test_direct.STRU")
        self.stru.write(stru_file, direct=True)
        self.assertTrue(os.path.exists(stru_file))

    def test_write_stru_cartesian(self):
        stru_file = os.path.join(self.work_dir, "test_cart.STRU")
        self.stru.write(stru_file, direct=False)
        self.assertTrue(os.path.exists(stru_file))

    def test_write2poscar(self):
        poscar_file = os.path.join(self.work_dir, "POSCAR")
        self.stru.write2poscar(poscar_file)
        self.assertTrue(os.path.exists(poscar_file))

    def test_to_ase(self):
        atoms = self.stru.to_ase()
        self.assertEqual(len(atoms), 2)

    def test_to_ase_empty2x(self):
        self.stru.set_empty_atom("Fe")
        atoms = self.stru.to_ase(empty2x=True)
        self.assertEqual(len(atoms), 2)

    def test_split_list(self):
        alist = [1, 2, 3, 4, 5]
        indices = [2, 3]
        result = self.stru.split_list(alist, indices)
        self.assertEqual(result, [[1, 2], [3, 4, 5]])

    def test_set_coord_cartesian(self):
        new_coord = [[0.5, 0.5, 0.5], [1.0, 1.0, 1.0]]
        self.stru.set_coord(new_coord, direct=False, bohr=False)
        coord = self.stru.get_coord(bohr=False, direct=False)
        self.assertEqual(len(coord), 2)

    def test_set_coord_direct(self):
        new_coord = [[0, 0, 0], [0.5, 0.5, 0.5]]
        self.stru.set_coord(new_coord, direct=True, bohr=False)
        coord = self.stru.get_coord(bohr=False, direct=True)
        self.assertEqual(len(coord), 2)

    def test_set_cell(self):
        new_cell = [[3.0, 0, 0], [0, 3.0, 0], [0, 0, 3.0]]
        self.stru.set_cell(new_cell, bohr=False, change_coord=True)
        cell = self.stru.get_cell(bohr=False)
        self.assertEqual(len(cell), 3)

    def test_get_stru(self):
        stru_dict = self.stru.get_stru()
        self.assertIn("cell", stru_dict)
        self.assertIn("coord", stru_dict)

    def test_perturb_stru_no_perturb(self):
        new_stru_list = self.stru.perturb_stru(0)
        self.assertEqual(len(new_stru_list), 1)

    def test_perturb_stru_cell(self):
        new_stru_list = self.stru.perturb_stru(2, cell_pert_frac=0.1)
        self.assertEqual(len(new_stru_list), 2)

    def test_perturb_stru_atom(self):
        new_stru_list = self.stru.perturb_stru(2, atom_pert_dist=0.1)
        self.assertEqual(len(new_stru_list), 2)

    def test_check_passes(self):
        self.assertTrue(self.stru._check())

    def test_get_dpks_none(self):
        dpks = self.stru.get_dpks()
        self.assertIsNone(dpks)


class TestAbacusStruMagToAngle(unittest.TestCase):
    """Test mag_to_angle and angle_to_mag methods"""

    def setUp(self):
        self.stru = abacus.AbacusStru(
            label=["Fe"],
            cell=[[2.87, 0, 0], [0, 2.87, 0], [0, 0, 2.87]],
            coord=[[0, 0, 0]],
            element=["Fe"],
        )

    def test_mag_to_angle_z_direction(self):
        angle1, angle2 = self.stru.mag_to_angle(0.0, 0.0, 1.0)
        self.assertAlmostEqual(angle1, 0.0, places=5)
        self.assertAlmostEqual(angle2, 0.0, places=5)

    def test_mag_to_angle_x_direction(self):
        angle1, angle2 = self.stru.mag_to_angle(1.0, 0.0, 0.0)
        self.assertAlmostEqual(angle1, 90.0, places=5)

    def test_angle_to_mag_z(self):
        mag = self.stru.angle_to_mag(1.0, 0.0, 0.0)
        self.assertAlmostEqual(mag[2], 1.0, places=5)

    def test_angle_to_mag_xy(self):
        mag = self.stru.angle_to_mag(1.0, 90.0, 0.0)
        self.assertAlmostEqual(mag[0], 1.0, places=5)


class TestAbacusStruMultipleAtomTypes(unittest.TestCase):
    """Test AbacusStru with multiple atom types"""

    def setUp(self):
        self.stru = abacus.AbacusStru(
            label=["Fe", "O", "Fe"],
            cell=[[10.0, 0, 0], [0, 10.0, 0], [0, 0, 10.0]],
            coord=[
                [0, 0, 0],
                [5, 5, 5],
                [2.5, 2.5, 2.5],
                [7.5, 7.5, 7.5],
                [3, 3, 3],
                [6, 6, 6],
            ],
            atom_number=[1, 2, 3],
            pp=["Fe.upf", "O.upf", "Fe.upf"],
            element=["Fe", "O", "Fe"],
        )

    def test_get_natoms(self):
        self.assertEqual(self.stru.get_natoms(), 6)

    def test_get_label_total(self):
        labels = self.stru.get_label(total=True)
        self.assertEqual(len(labels), 6)

    def test_get_element_total(self):
        elements = self.stru.get_element(total=True, number=False)
        self.assertEqual(elements.count("Fe"), 4)
        self.assertEqual(elements.count("O"), 2)

    def test_delete_one_type(self):
        new_stru = self.stru.delete_atom("O")
        self.assertEqual(new_stru.get_natoms(), 4)


class TestAbacusStruDirectCoordinates(unittest.TestCase):
    """Test AbacusStru with direct coordinates"""

    def setUp(self):
        self.stru = abacus.AbacusStru(
            label=["Fe"],
            cell=[[2.87, 0, 0], [0, 2.87, 0], [0, 0, 2.87]],
            coord=[[0, 0, 0], [0.5, 0.5, 0.5]],
            atom_number=[2],
            element=["Fe"],
            cartesian=False,
            lattice_constant=1.0,
        )

    def test_get_coord_direct(self):
        coord = self.stru.get_coord(bohr=False, direct=True)
        self.assertEqual(len(coord), 2)

    def test_get_coord_cartesian_from_direct(self):
        coord = self.stru.get_coord(bohr=False, direct=False)
        self.assertEqual(len(coord), 2)


if __name__ == "__main__":
    unittest.main()



import unittest,os,shutil,re
from pathlib import Path

from abacustest import prepare
import defaultset

class TestPrepare(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        self.work_path = Path("testprepare")
        os.makedirs(self.work_path,exist_ok=True)
        os.chdir(self.work_path)
        
        # create example template
        examples = ["a","b","c"]
        for i in examples:
            os.makedirs(i,exist_ok=True)
            Path(os.path.join(i,"STRU")).write_text(defaultset.STRU1)
            Path(os.path.join(i,"INPUT")).write_text(defaultset.INPUT1)
            Path(os.path.join(i,"Si.upf")).write_text("Si.upf")
            Path(os.path.join(i,"Si.orb")).write_text("Si.orb")
            Path(os.path.join(i,"KPT")).write_text(defaultset.KPT1)
        
        # create pp lib
        os.makedirs("pplib",exist_ok=True)
        Path("pplib/Si.upf1").write_text("Si.upf1")
        
        # create orb lib
        os.makedirs("orblib",exist_ok=True)
        Path("orblib/Si.orb1").write_text("Si.orb1")
        
        # create INPUT/KPT/STRU template
        Path("INPUT").write_text(defaultset.INPUT2)
        Path("KPT").write_text(defaultset.KPT2)
        Path("STRU").write_text(defaultset.STRU2)

    def tearDown(self):
        os.chdir(self.cwd)
        if os.path.isdir(self.work_path):
            shutil.rmtree(self.work_path)
    
    
    def test_prepare_example_template(self):
        #case1: use example template, and mix input
        param_setting = {
            "example_template": ["a","b","c"],
            "mix_input": {
                "ecutwfc": [50, 60, 70],
                "kspacing": [0.1, 0.12, 0.13]
            }
        }
        all_path_setting = prepare.DoPrepare(param_setting,"abacustest")
        
        # check result folder
        self.assertTrue(os.path.isdir("abacustest"))
        
        # check new create examples, should create 9 examples for each template
        self.assertEqual(len(all_path_setting), 3)
        self.assertEqual(len(all_path_setting[0]), 9)
        self.assertEqual(len(all_path_setting[1]), 9)
        self.assertEqual(len(all_path_setting[2]), 9)
        
        # check files in each example, and ecutwfc/kspacing value in INPUT
        ecut_kspacing_ref = [[50,0.1],[50,0.12],[50,0.13],[60,0.1],[60,0.12],[60,0.13],[70,0.1],[70,0.12],[70,0.13]]
        ecut_kspacing_ref.sort()
        for i in all_path_setting:
            ecut_kspacing = []
            for j in i:
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.upf")))
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.orb")))
                self.assertTrue(os.path.isfile(os.path.join(j,"STRU")))
                self.assertTrue(os.path.isfile(os.path.join(j,"INPUT")))
                self.assertTrue(os.path.isfile(os.path.join(j,"KPT")))
                input_text = Path(os.path.join(j,"INPUT")).read_text()
                ecutwfc = re.search(r"ecutwfc\s+[5,6,7]0",input_text)
                kspacing = re.search(r"kspacing\s+0\.1[0,2,3]?",input_text)
                self.assertTrue(ecutwfc)
                self.assertTrue(kspacing)
                ecut_kspacing.append([float(ecutwfc.group().split()[1]),float(kspacing.group().split()[1])])
            ecut_kspacing.sort()
            self.assertEqual(ecut_kspacing,ecut_kspacing_ref)
    
    def test_prepare_input_template(self):
        #case2: use INPUT template, and mix input/KPT 
        # check mix_kpt by using list or int
        param_setting = {
            "example_template": ["a","b","c"],
            "input_template": "INPUT",
            "mix_input": {
                "ecutwfc": [50, 60, 70]
            },
            "mix_kpt": [[1,2,3],2]
        }
        all_path_setting = prepare.DoPrepare(param_setting,"abacustest")  
        
        # check result folder
        self.assertTrue(os.path.isdir("abacustest"))
        
        # check new create examples, should create 2*3 = 6 examples for each template
        self.assertEqual(len(all_path_setting), 3)
        self.assertEqual(len(all_path_setting[0]), 6)
        self.assertEqual(len(all_path_setting[1]), 6)
        self.assertEqual(len(all_path_setting[2]), 6)
        
        # check files in each example, and ecutwfc/kspacing value in INPUT
        ecut_kpt_ref = [[50,[1,2,3]],[50,[2,2,2]],
                        [60,[1,2,3]],[60,[2,2,2]],
                        [70,[1,2,3]],[70,[2,2,2]]]
        
        for i in all_path_setting:
            ecut_kpt = []
            for j in i:
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.upf")))
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.orb")))
                self.assertTrue(os.path.isfile(os.path.join(j,"STRU")))
                self.assertTrue(os.path.isfile(os.path.join(j,"INPUT")))
                self.assertTrue(os.path.isfile(os.path.join(j,"KPT")))
                
                input_text = Path(os.path.join(j,"INPUT")).read_text()
                self.assertTrue(re.search(r"basis_type\s+pw",input_text))
                ecutwfc = re.search(r"ecutwfc\s+[5,6,7]0",input_text)
                self.assertTrue(ecutwfc)
                
                kpt_text = Path(os.path.join(j,"KPT")).read_text()
                kpt = [ int(i) for i in kpt_text.strip().split("\n")[-1].split()[:3]]
                ecut_kpt.append([float(ecutwfc.group().split()[1]),kpt])
            ecut_kpt.sort()
            self.assertEqual(ecut_kpt,ecut_kpt_ref) 

    def test_prepare_kpt_template(self):
        # case3: use kpt template, and mix STRU 
        param_setting = {
            "example_template": ["a","b","c"],
            "kpt_template": "KPT",
            "mix_stru": ["STRU","a/STRU"]
        }
        all_path_setting = prepare.DoPrepare(param_setting,"abacustest")  
        
        # check result folder
        self.assertTrue(os.path.isdir("abacustest"))
        
        # check new create examples, should create 2*3 = 6 examples for each template
        print(all_path_setting)
        self.assertEqual(len(all_path_setting), 3)
        self.assertEqual(len(all_path_setting[0]), 2)
        self.assertEqual(len(all_path_setting[1]), 2)
        self.assertEqual(len(all_path_setting[2]), 2)

        lc_ref = [5.1,10.2]
        for i in all_path_setting:
            lcs = []
            for j in i:
                self.assertTrue((os.path.isfile(os.path.join(j,"Si.upf")) or os.path.isfile(os.path.join(j,"Si.upf1"))))
                self.assertTrue((os.path.isfile(os.path.join(j,"Si.orb")) or os.path.isfile(os.path.join(j,"Si.orb1"))))
                self.assertTrue(os.path.isfile(os.path.join(j,"STRU")))
                self.assertTrue(os.path.isfile(os.path.join(j,"INPUT")))
                self.assertTrue(os.path.isfile(os.path.join(j,"KPT")))

                kpt_text = Path(os.path.join(j,"KPT")).read_text()
                kpt = [ int(i) for i in kpt_text.strip().split("\n")[-1].split()[:3]]
                self.assertEqual(kpt,[2,2,2]) 
                
                stru_text = Path(os.path.join(j,"STRU")).read_text().split("\n")
                #read LATTICE_CONSTANT, the value is in the next line of key word "LATTICE_CONSTANT"
                lc = 0
                for i,line in enumerate(stru_text):
                    if re.search(r"LATTICE_CONSTANT",line):
                        lc = float(stru_text[i+1].strip().split()[0])
                        break
                self.assertTrue(lc in lc_ref)
                lcs.append(lc)
            lcs.sort()
            self.assertEqual(lcs,lc_ref)
    
    def test_prepare_stru_template(self):
        # case4: use STRU template, and USE PP/ORB LIB
        # when specify pp_path or orb_path, prepare will firstly search pporb in lib
        # and if there is no pporb, then use pporb defined STRU
        param_setting = {
            "example_template": ["a","b","c"],
            "stru_template": "a/STRU",
            "pp_path": "pplib",
            "orb_path": "orblib"
        }
        all_path_setting = prepare.DoPrepare(param_setting,"abacustest")  
        
        # check result folder
        self.assertTrue(os.path.isdir("abacustest"))
        
        # check new create examples, should create 3 examples for each template
        print(all_path_setting)
        self.assertEqual(len(all_path_setting), 3)
        self.assertEqual(len(all_path_setting[0]), 1)
        self.assertEqual(len(all_path_setting[1]), 1)
        self.assertEqual(len(all_path_setting[2]), 1)

        for i in all_path_setting:
            for j in i:
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.upf1")))
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.orb1")))
                self.assertTrue(os.path.isfile(os.path.join(j,"STRU")))
                self.assertTrue(os.path.isfile(os.path.join(j,"INPUT")))
                self.assertTrue(os.path.isfile(os.path.join(j,"KPT")))
                
                self.assertEqual(Path(os.path.join(j,"Si.upf1")).read_text(),"Si.upf1")
                self.assertEqual(Path(os.path.join(j,"Si.orb1")).read_text(),"Si.orb1")
                stru_text = Path(os.path.join(j,"STRU")).read_text().split("\n")
                #read LATTICE_CONSTANT, the value is in the next line of key word "LATTICE_CONSTANT"
                lc = 0
                pp = orb = ""
                for i,line in enumerate(stru_text):
                    if re.search(r"LATTICE_CONSTANT",line):
                        lc = float(stru_text[i+1].strip().split()[0])
                    elif re.search(r"ATOMIC_SPECIES",line):
                        pp = stru_text[i+1].strip().split()[2]
                    elif re.search(r"NUMERICAL_ORBITAL",line):
                        orb = stru_text[i+1].strip().split()[0]
            
                self.assertEqual(lc, 10.2)
                self.assertTrue(pp in ["Si.upf1","./Si.upf1"])
                self.assertTrue(orb in ["Si.orb1","./Si.orb1"])
    
    def test_prepare_specify_pporb(self):
        # case4: use kpt template, and USE PP/ORB LIB
        param_setting = {
            "example_template": ["a","b","c"],
            "pp_dict": {"Si": "pplib/Si.upf1"},
            "orb_dict": {"Si": "a/Si.orb"}
        }
        all_path_setting = prepare.DoPrepare(param_setting,"abacustest")  
        
        # check result folder
        self.assertTrue(os.path.isdir("abacustest"))
        
        # check new create examples, should create 3 examples for each template
        print(all_path_setting)
        self.assertEqual(len(all_path_setting), 3)
        self.assertEqual(len(all_path_setting[0]), 1)
        self.assertEqual(len(all_path_setting[1]), 1)
        self.assertEqual(len(all_path_setting[2]), 1)

        for i in all_path_setting:
            for j in i:
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.upf1")))
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.orb")))
                self.assertTrue(os.path.isfile(os.path.join(j,"STRU")))
                self.assertTrue(os.path.isfile(os.path.join(j,"INPUT")))
                self.assertTrue(os.path.isfile(os.path.join(j,"KPT")))
                
                self.assertEqual(Path(os.path.join(j,"Si.upf1")).read_text(),"Si.upf1")
                self.assertEqual(Path(os.path.join(j,"Si.orb")).read_text(),"Si.orb")
                
                stru_text = Path(os.path.join(j,"STRU")).read_text().split("\n")
                #read LATTICE_CONSTANT, the value is in the next line of key word "LATTICE_CONSTANT"
                lc = 0
                pp = orb = ""
                for i,line in enumerate(stru_text):
                    if re.search(r"LATTICE_CONSTANT",line):
                        lc = float(stru_text[i+1].strip().split()[0])
                    elif re.search(r"ATOMIC_SPECIES",line):
                        pp = stru_text[i+1].strip().split()[2]
                    elif re.search(r"NUMERICAL_ORBITAL",line):
                        orb = stru_text[i+1].strip().split()[0]
            
                self.assertEqual(lc, 10.2)
                self.assertTrue(pp in ["Si.upf1","./Si.upf1"])
                self.assertTrue(orb in ["Si.orb","./Si.orb"])
    
    def test_prepare_dpks(self):
        # case4: use kpt template, and USE PP/ORB LIB
        Path("INPUT").write_text("INPUT_PARAMETERS\nbasis_type lcao\ndeepks_scf 1\ndeepks_model dpks.ptg")
        Path("dpks.ptg").write_text("dpks.ptg")
        Path("jle.orb").write_text("jle.orb")
        param_setting = {
            "example_template": ["a","b","c"],
            "input_template": "INPUT",
            "dpks_descriptor": "jle.orb",
            "extra_files": ["dpks.ptg"]
        }
        all_path_setting = prepare.DoPrepare(param_setting,"abacustest")  
        
        # check result folder
        self.assertTrue(os.path.isdir("abacustest"))
        
        # check new create examples, should create 3 examples for each template
        print(all_path_setting)
        self.assertEqual(len(all_path_setting), 3)
        self.assertEqual(len(all_path_setting[0]), 1)
        self.assertEqual(len(all_path_setting[1]), 1)
        self.assertEqual(len(all_path_setting[2]), 1)

        for i in all_path_setting:
            for j in i:
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.upf")))
                self.assertTrue(os.path.isfile(os.path.join(j,"Si.orb")))
                self.assertTrue(os.path.isfile(os.path.join(j,"STRU")))
                self.assertTrue(os.path.isfile(os.path.join(j,"INPUT")))
                self.assertTrue(os.path.isfile(os.path.join(j,"KPT")))
                self.assertTrue(os.path.isfile(os.path.join(j,"dpks.ptg")))
                self.assertTrue(os.path.isfile(os.path.join(j,"jle.orb")))
                
                stru_text = Path(os.path.join(j,"STRU")).read_text().split("\n")
                #read LATTICE_CONSTANT, the value is in the next line of key word "LATTICE_CONSTANT"
                dpks_descriptor = ""
                for i,line in enumerate(stru_text):
                    if re.search(r"NUMERICAL_DESCRIPTOR",line):
                        dpks_descriptor = stru_text[i+1].strip().split()[0]
                        break
                self.assertTrue(dpks_descriptor in ["jle.orb","./jle.orb"])       
    